# -----------------------------------------------------------------------------
# Copyright (C) 2004-2022 Gordon Smyth, Yifang Hu, Matthew Ritchie, Jeremy Silver, James Wettenhall, Davis McCarthy, Di Wu, Wei Shi, Belinda Phipson, Aaron Lun, Natalie Thorne, Alicia Oshlack, Carolyn de Graaf, Yunshun Chen, Mette Langaas, Egil Ferkingstad, Marcus Davy, Francois Pepin, Dongseok Choi
# Copyright (C) 2024 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# This file is based on the file 'R/dups.R' of the Bioconductor limma package (version 3.55.1).

import numpy as np


def unwrapdups(M, ndups=2, spacing=1):
    """
    Unwraps a matrix (M) for a series of experiments, combining duplicate spots for each gene into a single row.

    Arguments
    ---------
    M : ndarray
        the input matrix containing data for multiple experiments.
    ndups : int, optional
        the number of duplicate spots for each gene. Defaults to 2.
    spacing : int, optional
        the number of rows between duplicate measurements for a gene. Defaults to 1.

    Returns
    -------
    ndarray
        the unwrapped matrix with duplicate spots combined.
    """
    if ndups <= 0:
        raise ValueError("ndups must be strictly positive")
    if ndups == 1:
        return M

    M = np.asarray(M)  # Convert to NumPy array
    if M.ndim == 1:
        M = M[:, None]
    nspots, nslides = M.shape  # Get number of spots and slides

    ngroups = nspots // (ndups * spacing)
    new_shape = (spacing, ndups, ngroups, nslides)
    M = M.reshape(new_shape, order="F")

    # Transpose to combine duplicate spots for each gene
    M = np.transpose(M, (0, 2, 1, 3))
    return M.reshape(
        spacing * ngroups, ndups * nslides, order="F"
    )  # Reshape final output


def uniquegenelist(genelist, ndups=2, spacing=1):
    """
    Eliminates entries in a gene list for duplicate spots based on unwrapped data.

    Arguments
    ---------
    genelist : array_like
        The list of genes with potentially duplicate spots.
    ndups : int, optional
        The number of duplicate spots for each gene. Defaults to 2.
    spacing : int, optional
        The number of rows between duplicate measurements for a gene. Defaults to 1.

    Returns
    -------
    list
        The list of genes with duplicate spots removed.
    """
    if ndups <= 1:
        return genelist

    i = np.squeeze(
        unwrapdups(np.arange(genelist.shape[0]), ndups=ndups, spacing=spacing)[:, 0]
    )
    if i.ndim == 0:
        i = i.reshape(1)
    if genelist.ndim == 1:
        return genelist[i]
    else:
        return genelist[i, :]



# -----------------------------------------------------------------------------
# Copyright (C) 2004-2022 Gordon Smyth, Yifang Hu, Matthew Ritchie, Jeremy Silver, James Wettenhall, Davis McCarthy, Di Wu, Wei Shi, Belinda Phipson, Aaron Lun, Natalie Thorne, Alicia Oshlack, Carolyn de Graaf, Yunshun Chen, Mette Langaas, Egil Ferkingstad, Marcus Davy, Francois Pepin, Dongseok Choi
# Copyright (C) 2024 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

import numpy as np


class MArrayLM:
    """
    A class to store the results of fitting gene-wise linear models to a set of microarrays

    Objects of this class are normally created by :func:`lmFit` and additional
    components are added by :func:`eBayes`.

    Attributes
    ----------
    coefficients : pd.DataFrame
        matrix containing fitted coefficients or contrasts
    stdev_unscaled : pd.DataFrame
        matrix containing unscaled standard deviations of the coefficients or contrasts
    sigma : ndarray
        array containing residual standard deviations for each gene
    df_residual : ndarray
        array containing residual degrees of freedom for each gene
    Amean : ndarray, optional
        array containing the average log-intensity for each probe over all the
        arrays in the original linear model fit. Note that this vector does not
        change when a contrast is applied to the fit using
        :func:`contrasts_fit`.
    genes : pd.DataFrame, optional
        data frame containing probe annotation
    design : patsy.DesignMatrix, optional
        design matrix
    cov_coefficients : pd.DataFrame, optional
        matrix giving the unscaled covariance matrix of the estimable coefficients
    contrasts : pd.DataFrame, optional
        matrix defining contrasts of coefficients for which results are desired

    s2_prior : ndarray, optional
        single value or array giving empirical Bayes estimated prior value for
        residual variances
    df_prior : ndarray, optional
        value or vector giving empirical Bayes estimated degrees of freedom
        associated with :code:`s2_prior` for each gene
    df_total : ndarray, optional
        array giving total degrees of freedom used for each gene, usually equal
        to :code:`df_prior + df_residual`
    s2_post : ndarray, optional
        array giving posterior residual variances
    var_prior : ndarray, optional
        array giving empirical Bayes estimated prior variance for each true coefficient
    F : ndarray, optional
        array giving moderated *F*-statistics for testing all contrasts equal to zero
    F_p_value : ndarray, optional
        array giving *p*-value corresponding to :code:`F_stat`
    t : pd.DataFrame, optional
        matrix containing empirical Bayes *t*-statistics
    p_value : pd.DataFrame, optional
        matrix of two-sided *p*-values corresponding to the *t*-statistics
    lods : pd.DataFrame, optional
        matrix giving the log-odds of differential expression (on the natural log scale)
    """

    def __init__(self, coefficients, stdev_unscaled, sigma, df_residual, cov_coef):
        self.coefficients = coefficients
        self.stdev_unscaled = stdev_unscaled
        self.sigma = sigma
        self.df_residual = df_residual
        self.cov_coefficients = cov_coef

        self.Amean = None
        self.genes = None
        self.design = None
        self.contrasts = None
        self.s2_prior = None
        self.df_prior = None
        self.df_total = None
        self.s2_post = None
        self.var_prior = None
        self.F = None
        self.F_p_value = None
        self.t = None
        self.p_value = None
        self.lods = None

    def __getitem__(self, idx):
        row_idx, col_idx = idx

        res = MArrayLM(None, None, None, None, None)
        res.coefficients = self.coefficients.loc[row_idx, col_idx]
        res.stdev_unscaled = self.stdev_unscaled.loc[row_idx, col_idx]
        if self.sigma is not None:
            res.sigma = self.sigma[row_idx]
        if self.df_residual is not None:
            res.df_residual = self.df_residual[row_idx]
        if self.cov_coefficients is not None:
            res.cov_coefficients = self.cov_coefficients.loc[col_idx, col_idx]

        if self.Amean is not None:
            res.Amean = self.Amean[row_idx]
        if self.genes is not None:
            res.genes = self.genes.loc[row_idx]
        res.design = self.design
        if self.contrasts is not None:
            res.contrasts = self.contrasts[col_idx]
        if self.s2_prior is not None:
            if not isinstance(self.s2_prior, np.ndarray):
                res.s2_prior = self.s2_prior
            else:
                res.s2_prior = self.s2_prior[row_idx]
        if self.df_prior is not None:
            if not isinstance(self.df_prior, np.ndarray):
                res.df_prior = self.df_prior
            else:
                res.df_prior = self.df_prior[row_idx]
        if self.df_total is not None:
            res.df_total = self.df_total[row_idx]
        if self.s2_post is not None:
            res.s2_post = self.s2_post[row_idx]
        if self.var_prior is not None:
            res.var_prior = self.var_prior[row_idx]
        if self.F is not None:
            res.F = self.F[row_idx]
        if self.F_p_value is not None:
            res.F_p_value = self.F_p_value[row_idx]
        if self.t is not None:
            res.t = self.t.loc[row_idx, col_idx]
        if self.p_value is not None:
            res.p_value = self.p_value.loc[row_idx, col_idx]
        if self.lods is not None:
            res.lods = self.lods.loc[row_idx, col_idx]

        return res



# -----------------------------------------------------------------------------
# Copyright (C) 2024 M. Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

import pandas as pd


class DEResults(pd.DataFrame):
    """
    A class to store the results of differential expression analysis

    It serves as a common class to store the results of differential
    expression analysis from DESeq2, edgeR, limma.
    """

    from .plot import plotMA_res as plotMA

    @property
    def _constructor(self):
        return DEResults

    @property
    def _constructor_sliced(self):
        return pd.Series

    def __init__(self, df, *args, **kwargs):
        super().__init__(df, *args, **kwargs)
        # TODO what about baseMean (mean of normalized counts)?
        # TODO what about adjusted p-values (would be adj_pvalue)?
        for col in ["log2FoldChange", "lfcSE", "pvalue"]:
            if col not in self.columns:
                raise ValueError(f"{col} missing from results table")



from .factor import Factor as Factor
from .factor import asfactor as asfactor
from .factor import gl as gl
from .format import round_scientific_notation as round_scientific_notation
from .format import truncate_name as truncate_name
from .linalg import cov2cor as cov2cor
from .lm import lm_fit as lm_fit
from .lm import lm_wfit as lm_wfit
from .logging import LOGGER as LOGGER
from .splines import ns as ns
from .splines import spline_design as spline_design
from .stats import dnbinom_mu as dnbinom_mu
from .stats import dnorm as dnorm
from .stats import pnorm as pnorm
from .stats import pt as pt
from .stats import rnbinom as rnbinom



# -----------------------------------------------------------------------------
# Copyright (C) 2024 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

import numpy as np
from statsmodels.regression.linear_model import WLS


class lm_wfit:
    """
    Python equivalent to R :code:`lm.wfit`

    This class is merely a wrapper around :class:`RegressionResults`, the
    return type of :meth:`OLS.fit` and :meth:`WLS.fit`.

    It contains the solution to the equation :math:`A X = B`, where :math:`A`
    is a known matrix, :math:`B` a known vector and :math:`X` the unknown to
    solve for.  The solution is obtained by a weighted least-square-error fit
    based on the QR-decomposition of :math:`A`.

    Arguments
    ---------
    design : array_like
        matrix :math:`A` of the equation
    b : array_like
        offset (vector :math:`B` in the equation)
    w : array_like
        weights to use in the least-square-error fit
    """

    def __init__(self, design, b, w):
        self.fit = WLS(b, design, w).fit(method="qr")
        self._effects = None

    @property
    def coefficients(self):
        """
        ndarray: the value found for :math:`X`
        """
        return self.fit.params

    @property
    def residuals(self):
        """
        ndarray: the residuals, *i.e.* the difference between :math:`A X` and :math:`B`
        """
        return self.fit.resid

    @property
    def fitted_values(self):
        """
        ndarray: the value :math:`A X`
        """
        return self.fit.fittedvalues

    @property
    def effects(self):
        """
        ndarray: vector of orthogonal single-df effects. The first :code:`rank` of them
        correspond to the :code:`coefficients`
        """
        if self._effects is None:
            Q, _ = np.linalg.qr(self.fit.model.wexog, mode="complete")
            self._effects = Q.T @ self.fit.model.wendog
        return self._effects

    @property
    def weights(self):
        """
        ndarray: the weights used in the least-square-error fit
        """
        return self.fit.weights

    @property
    def rank(self):
        """
        int: the column rank of matrix :math:`A`
        """
        return self.fit.model.rank

    @property
    def df_residuals(self):
        """
        ndarray: degrees of freedom of residuals
        """
        return self.fit.model.df_resid

    @property
    def qr(self):
        """
        ndarray, ndarray: the QR decomposition of matrix :math:`A`
        """
        return self.fit.model.exog_Q, self.fit.model.exog_R


class lm_fit(lm_wfit):
    """
    Python equivalent to R :code:`lm.fit`

    This class is merely a wrapper around :class:`RegressionResults`, the
    return type of :meth:`OLS.fit` and :meth:`WLS.fit`.

    It contains the solution to the equation :math:`A X = B`, where :math:`A`
    is a known matrix, :math:`B` a known vector and :math:`X` the unknown to
    solve for.  The solution is obtained by a least-square-error fit based on
    the QR-decomposition of :math:`A`.

    Arguments
    ---------
    design : array_like
        matrix :math:`A` of the equation
    b : array_like
        offset (vector :math:`B` in the equation)
    """

    def __init__(self, design, b):
        super().__init__(design, b, 1.0)



# -----------------------------------------------------------------------------
# Copyright (C) 2008-2022 Yunshun Chen, Aaron TL Lun, Davis J McCarthy, Matthew E Ritchie, Belinda Phipson, Yifang Hu, Xiaobei Zhou, Mark D Robinson, Gordon K Smyth
# Copyright (C) 2024-2025 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# This file is based on the files 'R/topTags.R' of the Bioconductor edgeR package (version 3.38.4).

import numpy as np
import pandas as pd
from statsmodels.stats.multitest import multipletests

from ..diffexp import DEResults
from ..utils import LOGGER
from .DGEExact import DGEExact


class TopTags(DEResults):
    _metadata = ["adjust_method", "comparison", "test"]

    @property
    def _constructor(self):
        def f(*args, **kwargs):
            return TopTags(
                *args,
                adjust_method=self.adjust_method,
                comparison=self.comparison,
                test=self.test,
            )

        return f

    @property
    def _constructor_slided(self):
        return pd.Series

    def __init__(self, table, adjust_method, comparison, test, *args, **kwargs):
        super().__init__(table, *args, **kwargs)
        self.adjust_method = adjust_method
        self.comparison = comparison
        self.test = test

    def __repr__(self):
        return f"Coefficient: {self.comparison}\n{self}"


def topTags(self, n=10, adjust_method="fdr_bh", sort_by="PValue", p_value=1):
    """
    Extract the most differentially expressed genes (or sequence tags) from a test object, ranked either by *p*-value or by absolute log-fold-change.

    This function accepts a test statistic object created by any of the
    functions :func:`exactTest`, :func:`glmLRT`, :func:`glmTreat` or
    :func:`glmQLFTest` and extracts a readable dataframe of the most
    differentially expressed genes.  The dataframe collates the annotation and
    differential expression statistics for the top genes. The dataframe is
    wrapped in a :class:`TopTags` object that records the test statistic used
    and the multiple testing adjustment method.

    :func:`topTags` permits ranking by fold-change but the authors do not
    recommend fold-change ranking or fold-change cutoffs for routine RNA-Seq
    analysis. The *p*-value ranking is intended to be more biologically
    meaningful, especially if the *p*-values were computed using
    :func:`glmTreat`.

    Arguments
    ---------
    self : DGEExact or DGELRT
        object containing test statistics and *p*-values
    n : int
        maximum number of genes/tags to return
    adjust_method : str
        specify the method used to adjust *p*-values for multiple testing. See
        :func:`statsmodels.stats.multitest` for possible values
    sort_by : {"PValue, "logFC", "none"}
        specify the sort method

        - :code:`"PValue"` to sort by *p*-value
        - :code:`"logFC"` to sort by absolute log-fold-change
        - :code:`"none"` for no sorting

    p_value : float
        cutoff value for adjusted *p*-values. Only tags with adjusted
        *p*-values equal or lower than specified are returned.

    Returns
    -------
    TopTags
        a dataframe containing differential expression results for the top
        genes in a sorted order. The number of rows is the smaller of :code:`n`
        and the number of genes with adjusted *p*-value less than or equal to
        :code:`p_value`. The dataframe includes all the annotation columns from
        :code:`self.genes` and all statistic columns from :code:`self` plus one
        of:

        - :code:`"FDR"`, false discovery rate (only when
          :code:`adjust_method` is :code:`"fdr_bh"`, :code:`"fdr_by"`))
        - :code:`"FWER"`, family-wise error rate (only when
          :code:`adjust_method` is :code:`"holm"`, :code:`"simes-hochberg"`,
          :code:`"hommel"` or :code:`"bonferroni"`)

        For consistency with other modules, the dataframe also contains a
        :code:`"adj_pvalue"` column with the same content as the :code:`"FDR"`
        or :code:`"FWER"` column.

        The object also contains the following components:

        - :code:`adjust_method`, string specifying the method used to adjust
          *p*-values for multiple testing, same as input argument
        - :code:`comparison` the names of the two groups being compared (for
          :class:`DGEExact` objects) or the glm contrast being tested (for
          :class:`DGELRT` objects).
        - :code:`test`, string stating the name of the test
    """
    if isinstance(self, DGEExact):
        test = "exact"
    else:
        test = "glm"
    MultipleContrasts = test == "glm" and self.shape[1] > 5

    # Check adjust_method
    FWER_methods = ["holm", "simes-hochberg", "hommel", "bonferroni"]
    FDR_methods = ["fdr_bh", "fdr_by"]
    if adjust_method not in FWER_methods and adjust_method not in FDR_methods:
        raise ValueError(f"invalid argument {adjust_method} for 'adjust_method'")

    # Check sort_by
    if sort_by == "p_value":
        sort_by = "PValue"
    if sort_by not in ["none", "PValue", "logFC"]:
        raise ValueError(f"invalid value {sort_by} for 'sort_by'")

    # Absolute log fold change
    if MultipleContrasts:
        if sort_by == "logFC":
            LOGGER.warning(
                "Two or more logFC columns in DGELRT object. First logFC column used to rank by logFC"
            )
        alfc = np.abs(self["log2FoldChange"].iloc[:, 0])
    else:
        alfc = np.abs(self["log2FoldChange"])

    # Choose top genes
    if sort_by == "logFC":
        o = np.argsort(-alfc)
    elif sort_by == "PValue":
        o = np.argsort(
            np.array(
                [(a, b) for a, b in zip(self["pvalue"], -alfc)],
                dtype=[("x", self["pvalue"].dtype), ("y", alfc.dtype)],
            ),
            order=("x", "y"),
        )
    elif sort_by == "none":
        o = np.arange(self.shape[0])
    else:
        raise ValueError(f"invalid value {sort_by} for 'sort_by'")
    tab = self.iloc[o, :]

    # Add adjusted p-values if appropriate
    adj_p_val = multipletests(self["pvalue"], method=adjust_method)[1]
    if adjust_method != "none":
        if adjust_method in FWER_methods:
            adjustment = "FWER"
        if adjust_method in FDR_methods:
            adjustment = "FDR"
        tab[adjustment] = adj_p_val[o]
        tab["adj_pvalue"] = adj_p_val[o]

    # Add gene annotation if appropriate
    if self.genes is not None:
        rn = self.genes.index
        for c in self.genes.columns:
            tab[c] = self.genes[c]
        tab.index = rn

    # Thin out fit p_value threshold
    if p_value < 1:
        sig = adj_p_val[o] <= p_value
        sig[np.isnan(sig)] = False
        tab = tab[sig, :]

    # Enough rows left?
    if tab.shape[0] < n:
        n = tab.shape[0]
    if n < 1:
        raise ValueError("No rows to output")

    # Output object
    return TopTags(
        table=tab.iloc[:n, :],
        adjust_method=adjust_method,
        comparison=self.comparison,
        test=test,
    )



import os
import unittest

import pandas as pd

from inmoose.cohort_qc.cohort_metric import CohortMetric
from inmoose.cohort_qc.qc_report import QCReport


class TestQCReport(unittest.TestCase):
    def setUp(self):
        import os

        self.output_path = os.path.join(
            os.path.dirname(__file__),
            "..",
            "html_report_output",
            "test_qc_report_output.html",
        )
        output_dir = os.path.dirname(self.output_path)
        os.makedirs(output_dir, exist_ok=True)

        self.clinical_df = pd.DataFrame(
            {
                "SampleID": ["S1", "S2", "S3", "S4"],
                "batch": ["B1", "B1", "B2", "B2"],
                "Cov1": [1, 1, 2, 2],
                "Cov2": ["X", "Y", "X", "Y"],
            }
        ).set_index("SampleID")

        self.expression_df = pd.DataFrame(
            {
                "S1": [1.1, 1.2, 1.3],
                "S2": [2.1, 2.2, 2.3],
                "S3": [3.1, 3.2, 3.3],
                "S4": [4.1, 4.2, 4.3],
            },
            index=["G1", "G2", "G3"],
        )

        self.qc = CohortMetric(
            clinical_df=self.clinical_df,
            batch_column="batch",
            data_expression_df=self.expression_df,
            data_expression_df_before=self.expression_df.copy(),
            covariates=["Cov1", "Cov2"],
            n_components=2,
            n_neighbors=2,
        )
        self.qc.process()

        self.report = QCReport(self.qc)
        self.report.save_report(output_path=self.output_path)

    def test_html_report_created(self):
        """Vérifie que le fichier HTML est bien créé."""
        self.assertTrue(os.path.isfile(self.output_path))

    def test_html_report_not_empty(self):
        """Vérifie que le fichier HTML généré n'est pas vide."""
        with open(self.output_path, "r", encoding="utf-8") as f:
            content = f.read()
        self.assertGreater(len(content.strip()), 100)

    def test_html_report_contains_expected_sections(self):
        """Vérifie que le HTML contient les titres de section attendus."""
        with open(self.output_path, "r", encoding="utf-8") as f:
            html = f.read()

        expected_sections = [
            "Cohort Summary",
            "PCA Analysis and Variance Explained",
            "Correction Effect Metric",
            "Silhouette Score",
            "Entropy of Batch Mixing",
            "Mixed Dataset Summary Report",
        ]

        for section in expected_sections:
            with self.subTest(section=section):
                self.assertIn(section, html)

    def test_html_contains_images(self):
        """Vérifie qu’il y a bien des images encodées base64 dans le HTML."""
        with open(self.output_path, "r", encoding="utf-8") as f:
            html = f.read()

        self.assertIn("data:image/png;base64,", html)


if __name__ == "__main__":
    unittest.main()



# -----------------------------------------------------------------------------
# Copyright (C) 2008-2022 Yunshun Chen, Aaron TL Lun, Davis J McCarthy, Matthew E Ritchie, Belinda Phipson, Yifang Hu, Xiaobei Zhou, Mark D Robinson, Gordon K Smyth
# Copyright (C) 2022-2023 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# This file is a Python port of the original C++ code from the files
# 'src/glm_one_group.cpp' and 'src/R_fit_one_group.cpp' of the Bioconductor
# edgeR package (version 3.38.4).

import numpy as np

from .edgepy_cpp import glm_one_group_cython


def glm_one_group(counts, offset, disp, weights, maxit, tolerance, cur_beta):
    """
    Simplified fit for negative binomial GLM when the design matrix is one group

    This is the low-level function called by :func:`fit_one_group`.

    See Also
    --------
    fit_one_group

    Arguments
    ---------
    counts : array_like
        matrix of counts
    offset : array_like
        matrix of offsets for the log-linear models
    disp : array_like
        matrix of dispersions
    weights : array_like
        matrix of observation weights
    maxit : int
        maximum number of Newton-Raphson iterations
    tol : float
        tolerance for convergence in the Newton-Raphson iteration
    cur_beta : array_like
        initial coefficients

    Returns
    -------
    ndarray
        fitted coefficients (one per row in :code:`y`)
    ndarray
        whether the fit converged for each coefficient
    """

    out_beta = cur_beta.copy()
    out_conv = np.repeat(False, out_beta.shape)

    for j in range(out_beta.shape[0]):
        (out_beta[j], out_conv[j]) = glm_one_group_cython(
            counts.shape[1],
            counts[j, :],
            offset[j, :],
            disp[j, :],
            weights[j, :],
            maxit,
            tolerance,
            cur_beta[j],
        )
    return (out_beta, out_conv)


def fit_one_group(y, offsets, disp, weights, maxit, tol, beta, usePoisson=True):
    """
    Simplified fit for negative binomial GLM when the design matrix is one group

    In essence, this function has the same purpose as :func:`fit_levenberg`,
    specialized for the case where the design matrix is one group. In
    particular, Levenberg damping is not required here.

    This function also uses a Poisson approximation for zero dispersion genes,
    and calls :func:`glm_one_group` otherwise.

    See Also
    --------
    fit_levenberg
    glm_one_group

    Arguments
    ---------
    y : array_like
        matrix of counts
    offsets : array_like
        matrix of offsets for the log-linear models
    disp : array_like
        matrix of dispersions
    weights : array_like
        matrix of observation weights
    maxit : int
        maximum number of Newton-Raphson iterations
    tol : float
        tolerance for convergence in the Newton-Raphson iteration
    beta : array_like
        initial coefficients
    usePoisson : bool, optional
        whether to use Poisson approximation for zero-disp one-weight genes.
        Defaults to :code:`True`

    Returns
    -------
    ndarray
        fitted coefficients (one per row in :code:`y`)
    ndarray
        whether the fit converged for each coefficient
    """

    assert beta.shape == (y.shape[0],)

    out_beta = beta.copy()
    out_conv = np.repeat(False, beta.shape)

    # Checking for the Poisson special case with all-unity weights and all-zero
    # dispersions
    if usePoisson:
        disp_is_zero = (disp == 0).all(axis=1)
        weight_is_one = (weights == 1).all(axis=1)
        special_case = disp_is_zero & weight_is_one
        general_case = ~special_case
    else:
        special_case = np.repeat(False, y.shape[0])
        general_case = ~special_case

    sum_lib = np.exp(offsets[special_case]).sum(axis=1)
    sum_counts = y[special_case].sum(axis=1)
    with np.errstate(divide="ignore"):
        out_beta[special_case] = np.where(
            sum_counts == 0, -np.inf, np.log(sum_counts / sum_lib)
        )
    out_conv[special_case] = True

    # Otherwise going through NR iterations
    out_beta[general_case], out_conv[general_case] = glm_one_group(
        y[general_case],
        offsets[general_case],
        disp[general_case],
        weights[general_case],
        maxit,
        tol,
        out_beta[general_case],
    )

    return (out_beta, out_conv)



import unittest

import numpy as np

from inmoose.limma import classifyTestsF


class Test(unittest.TestCase):
    def test_classifyTestsF(self):
        M = np.arange(1, 13).reshape((3, 4))
        self.assertTrue(
            np.array_equal(
                classifyTestsF(M), [[0, 0, 1, 1], [1, 1, 1, 1], [1, 1, 1, 1]]
            )
        )
        self.assertTrue(
            np.array_equal(
                classifyTestsF(M / 2), [[0, 0, 0, 0], [1, 1, 1, 1], [1, 1, 1, 1]]
            )
        )
        self.assertTrue(
            np.array_equal(
                classifyTestsF(M / 3), [[0, 0, 0, 0], [0, 1, 1, 1], [1, 1, 1, 1]]
            )
        )
        self.assertTrue(
            np.array_equal(
                classifyTestsF(M, df=3), [[0, 0, 0, 0], [0, 1, 1, 1], [1, 1, 1, 1]]
            )
        )
        self.assertTrue(
            np.array_equal(
                classifyTestsF(M, p_value=0.00001),
                [[0, 0, 0, 1], [1, 1, 1, 1], [1, 1, 1, 1]],
            )
        )

        fstat = classifyTestsF(M, fstat_only=True)
        self.assertTrue(np.array_equal(fstat, [7.5, 43.5, 111.5]))
        self.assertEqual(fstat.df1, 4)
        self.assertTrue(np.isposinf(fstat.df2))



import importlib.metadata

__version__ = "0.8.2-dev0" # importlib.metadata.version(__package__)

from . import edgepy as edgepy
from . import pycombat as pycombat
from . import utils as utils
from . import consensus_clustering as consensus_clustering
from . import deseq2 as deseq2
from . import diffexp as diffexp
from . import cohort_qc as cohort_qc



# distutils: language = c++
#-----------------------------------------------------------------------------
# Copyright (C) 2022-2025 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#-----------------------------------------------------------------------------

from libcpp cimport bool, pair, vector
from numpy cimport ndarray

# cf. file src/utils.h from edgeR source repo

cdef public ndarray[double, ndim=1] vector2ndarray "vector2ndarray"(const vector.vector[double]& data)
cdef public double compute_unit_nb_deviance(double, double, double)

cpdef public vector.vector[double] maximize_interpolant "maximize_interpolant"(vector.vector[double] spts, const double[:,:] likelihoods)



import logging

logging.captureWarnings(True)

LOGGER = logging.getLogger("inmoose")
handler = logging.StreamHandler()
formatter = logging.Formatter("[%(levelname)s] %(message)s")
handler.setFormatter(formatter)
LOGGER.handlers.clear()
LOGGER.addHandler(handler)



from .core import DESeq as DESeq
from .DESeqDataSet import DESeqDataSet as DESeqDataSet
from .DESeqDataSet import makeExampleDESeqDataSet as makeExampleDESeqDataSet
from .DESeqTransform import DESeqTransform as DESeqTransform
from .dispersions import (
    estimateDispersionsFit as estimateDispersionsFit,
)
from .dispersions import (
    estimateDispersionsGeneEst as estimateDispersionsGeneEst,
)
from .dispersions import (
    estimateDispersionsMAP as estimateDispersionsMAP,
)
from .dispersions import (
    estimateDispersionsPriorVar as estimateDispersionsPriorVar,
)
from .estimateSizeFactors import (
    estimateSizeFactorsForMatrix as estimateSizeFactorsForMatrix,
)
from .lrt import nbinomLRT as nbinomLRT
from .outliers import replaceOutliers as replaceOutliers
from .parallel import estimateMLEForBetaPriorVar as estimateMLEForBetaPriorVar
from .prior import estimateBetaPriorVar as estimateBetaPriorVar
from .replicates import collapseReplicates as collapseReplicates
from .vst import varianceStabilizingTransformation as varianceStabilizingTransformation
from .wald import nbinomWaldTest as nbinomWaldTest



# -----------------------------------------------------------------------------
# Copyright (C) 2008-2022 Yunshun Chen, Aaron TL Lun, Davis J McCarthy, Matthew E Ritchie, Belinda Phipson, Yifang Hu, Xiaobei Zhou, Mark D Robinson, Gordon K Smyth
# Copyright (C) 2024 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

import pandas as pd

from ..diffexp import DEResults


class DGEExact(DEResults):
    _metadata = ["comparison", "genes"]

    @property
    def _constructor(self):
        def f(*args, **kwargs):
            return DGEExact(
                *args, comparison=self.comparison, genes=self.genes, **kwargs
            )

        return f

    @property
    def _constructor_sliced(self):
        return pd.Series

    def __init__(self, table, comparison, genes, *args, **kwargs):
        super().__init__(table, *args, **kwargs)
        self.comparison = comparison
        self.genes = genes



# -----------------------------------------------------------------------------
# Copyright (C) 2008-2022 Yunshun Chen, Aaron TL Lun, Davis J McCarthy, Matthew E Ritchie, Belinda Phipson, Yifang Hu, Xiaobei Zhou, Mark D Robinson, Gordon K Smyth
# Copyright (C) 2022-2025 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# This file is based on the file 'R/glmfit.R' of the Bioconductor edgeR package (version 3.38.4).


import numpy as np
import pandas as pd
import scipy
from patsy import DesignMatrix, dmatrix

from ..utils import asfactor
from .DGEGLM import DGEGLM
from .DGELRT import DGELRT
from .makeCompressedMatrix import _compressDispersions, _compressOffsets
from .mglmLevenberg import mglmLevenberg
from .mglmOneWay import designAsFactor, mglmOneWay
from .nbinomDeviance import nbinomDeviance
from .predFC import predFC


def glmFit_DGEList(self, design=None, dispersion=None, prior_count=0.125, start=None):
    """
    Fit a negative binomial generalized log-linear model to the read counts for
    each gene. Conduct genewise statistical tests for a given coefficient or
    coefficient contrast.

    See also
    --------
    glmFit

    Arguments
    ---------
    design : matrix, optional
        design matrix for the genewise linear models. Must be of full column
        rank. Defaults to a single column of ones, equivalent to treating the
        columns as replicate libraries.
    dispersion : float or array_like
        scalar, vector or matrix of negative binomial dispersions. Can be a
        common value for all genese, a vector of dispersion values with one for
        each gene, or a matrix of dispersion values with one for each observation.
        If :code:`None`, it will be extracted from :code:`y`, with order of
        precedence: genewise dispersion, trended dispersion, common dispersion.
    prior_count : float
        average prior count to be added to observation to shrink the estimated
        log-fold-change towards zero.
    start : matrix, optional
        initial estimates for the linear model coefficients

    Returns
    -------
    DGEGLM
        object containing the data about the fit
    """

    # The design matrix defaults to the oneway layout defined by self.samples["group"]
    # If there is only one group, then the design matrix is left None so that a matrix with a single intercept column will be set later by glmFit.
    if design is None:
        design = self.design
        if design is None:
            group = asfactor(self.samples["group"]).droplevels()
            if group.nlevels() > 1:
                design = dmatrix("~C(self.samples['group'])")

    if dispersion is None:
        dispersion = self.getDispersion()
    if dispersion is None:
        raise ValueError("No dispersion values found in DGEList object")
    offset = self.getOffset()
    if self.AveLogCPM is None:
        self.AveLogCPM = self.aveLogCPM()

    fit = glmFit(
        y=self.counts,
        design=design,
        dispersion=dispersion,
        offset=offset,
        lib_size=None,
        weights=self.weights,
        prior_count=prior_count,
        start=start,
    )

    fit.samples = self.samples
    fit.genes = self.genes
    fit.prior_df = self.prior_df
    fit.AveLogCPM = self.AveLogCPM
    return fit


def glmFit(
    y,
    design=None,
    dispersion=None,
    offset=None,
    lib_size=None,
    weights=None,
    prior_count=0.125,
    start=None,
):
    """
    Fit a negative binomial generalized log-linear model to the read counts for
    each gene. Conduct genewise statistical tests for a given coefficient or
    coefficient contrast.

    This function implements one of the GLM methods developed by [McCarthy2012]_.

    :code:`glmFit` fits genewise negative binomial GLMs, all with the same
    design matrix but possibly different dispersions, offsets and weights.
    When the design matrix defines a one-way layout, or can be re-parameterized
    to a one-way layout, the GLMs are fitting very quickly using
    :func:`mglmOneGroup`. Otherwise the default fitting method, implemented in
    :func:`mglmLevenberg`, uses a Fisher scoring algorithm with Levenberg-style
    damping.

    Positive :code:`prior_count` cause the returned coefficients to be shrunk in
    such a way that fold-changes between the treatment conditions are decreased.
    In particular, infinite fold-changes are avoided. Larger values cause more
    shrinkage. The returned coefficients are affected but not the likelihood
    ratio tests or p-values.

    See also
    --------
    mglmOneGroup : low-level computations
    mglmLevenberg : low-level computations

    Arguments
    ---------
    y : pd.DataFrame
        matrix of counts
    design : matrix, optional
        design matrix for the genewise linear models. Must be of full column
        rank. Defaults to a single column of ones, equivalent to treating the
        columns as replicate libraries.
    dispersion : float or array_like
        scalar, vector or matrix of negative binomial dispersions. Can be a
        common value for all genes, a vector of dispersion values with one for
        each gene, or a matrix of dispersion values with one for each
        observation.
    offset : float or array_like, optional
        matrix of the same shape as :code:`y` giving offsets for the log-linear
        models. Can be a scalar or a vector of length :code:`y.shape[1]`, in
        which case it is broadcasted to the shape of :code:`y`.
    lib_size : array_like, optional
        vector of length :code:`y.shape[1]` giving library sizes. Only used if
        :code:`offset=None`, in which case :code:`offset` is set to
        :code:`log(lib_size)`. Defaults to :code:`colSums(y)`.
    weights : matrix, optional
        prior weights for the observations (for each library and gene) to be
        used in the GLM calculations
    prior_count : float
        average prior count to be added to observation to shrink the estimated
        log-fold-change towards zero.
    start : matrix, optional
        initial estimates for the linear model coefficients

    Returns
    -------
    DGEGLM
        object containing:

        - :code:`counts`, the input matrix of counts
        - :code:`design`, the input design matrix
        - :code:`weights`, the input weights matrix
        - :code:`offset`, matrix of linear model offsets
        - :code:`dispersion`, vector of dispersions used for the fit
        - :code:`coefficients`, matrix of estimated coefficients from the GLM
          fits, on the natural log scale, of size :code:`y.shape[0]` by
          :code:`design.shape[1]`.
        - :code:`unshrunk_coefficients`, matrix of estimated coefficients from
          the GLM fits when no log-fold-changes shrinkage is applied, on the
          natural log scale, of size :code:`y.shape[0]` by
          :code:`design.shape[1]`. It exists only when :code:`prior_count` is
          not 0.
        - :code:`fitted_values`, matrix of fitted values from GLM fits, same
          shape as :code:`y`
        - :code:`deviance`, numeric vector of deviances, one for each gene
    """
    # Check y
    (ntag, nlib) = y.shape

    # Check design
    if design is None:
        design = dmatrix("~1", pd.DataFrame(y.T))
    try:
        design = DesignMatrix(
            np.asarray(design, order="F"), design_info=design.design_info
        )
    except AttributeError:
        design = np.asarray(design, order="F")

    if design.shape[0] != nlib:
        raise ValueError("design should have as many rows as y has columns")
    if np.linalg.matrix_rank(design) < design.shape[1]:
        raise ValueError(
            "Design matrix is not full rank. Some coefficients are not estimable"
        )

    # Check dispersion
    if dispersion is None:
        raise ValueError("No dispersion values provided")
    dispersion = np.asanyarray(dispersion)
    # TODO check dispersion for NaN and non-numeric values
    if dispersion.shape not in [(), (1,), (ntag,), y.shape]:
        raise ValueError("Dimensions of dispersion do not agree with dimensions of y")
    dispersion_mat = _compressDispersions(y, dispersion)

    # Check offset
    if offset is not None:
        # TODO check that offset is numeric
        offset = np.asanyarray(offset)
        if offset.shape not in [(), (1,), (nlib,), y.shape]:
            raise ValueError("Dimensions of offset do not agree with dimensions of y")

    # Check lib_size
    if lib_size is not None:
        # TODO check that lib_size is numeric
        lib_size = np.asarray(lib_size)
        if lib_size.shape not in [(), (1,), (nlib,)]:
            raise ValueError("lib_size has wrong length, should agree with ncol(y)")

    # Consolidate lib_size and offset into a compressed matrix
    offset = _compressOffsets(y=y, lib_size=lib_size, offset=offset)

    # weights are checked in lower-level functions

    # Fit the tagwise GLMs
    # If the design is equivalent to a oneway layout, use a shortcut algorithm
    group = designAsFactor(design)
    if group.nlevels() == design.shape[1]:
        (coef, fitted_values) = mglmOneWay(
            y,
            design=design,
            group=group,
            dispersion=dispersion_mat,
            offset=offset,
            weights=weights,
            coef_start=start,
        )
        deviance = nbinomDeviance(
            y=y, mean=fitted_values, dispersion=dispersion_mat, weights=weights
        )
        fit_method = "oneway"
        fit = (coef, fitted_values, deviance, None, None)
    else:
        fit = mglmLevenberg(
            y,
            design=design,
            dispersion=dispersion_mat,
            offset=offset,
            weights=weights,
            coef_start=start,
            maxit=250,
        )
        fit_method = "levenberg"

    # Prepare output
    fit = DGEGLM(fit)
    fit.counts = y
    fit.method = fit_method
    if prior_count > 0:
        fit.unshrunk_coefficients = fit.coefficients
        fit.coefficients = predFC(
            y,
            design,
            offset=offset,
            dispersion=dispersion_mat,
            prior_count=prior_count,
            weights=weights,
        ) * np.log(2)

    # counts N,M
    # design M,P
    assert y.shape[1] == design.shape[0]
    w_vec = fit.fitted_values / (1.0 + dispersion_mat * fit.fitted_values)
    if weights is not None:
        w_vec = weights * w_vec
    ridge = np.diag(np.repeat(1e-6 / (np.log(2) ** 2), design.shape[1]))
    xtwxr_inv = np.linalg.inv(design.T @ (design * w_vec[:, :, None]) + ridge)
    sigma = xtwxr_inv @ design.T @ (design * w_vec[:, :, None]) @ xtwxr_inv
    fit.coeff_SE = np.diagonal(sigma, axis1=-2, axis2=-1)

    # FIXME (from original R source) we are not allowing missing values, so df.residual must be same for all tags
    fit.df_residual = np.full(ntag, nlib - design.shape[1])
    fit.design = design
    fit.offset = offset
    fit.dispersion = dispersion
    fit.weights = weights
    fit.prior_count = prior_count
    return fit


def glmLRT(glmfit, coef=None, contrast=None):
    """
    Conduct genewise statistical tests for a given coefficient or coefficient contrast.

    This function implements one of the GLM methods developed by [McCarthy2012]_.

    :func:`glmLRT` conducts likelihood ratio tests for one or more coefficients
    in the linear model. If :code:`coef` is used, the null hypothesis is that
    all the coefficients indicated by :code:`coef` are equal to zero. If
    :code:`contrast` is non-null, then the null hypothesis is that the
    specified contrasts of the coefficients are equal to zero. For example, a
    contrast of :code:`[0,1,-1]`, assuming there are three coefficients, would
    test the hypothesis that the second and third coefficients are equal.

    Arguments
    ---------
    glmfit : DGEGLM
        a :class:`DGEGLM` object, usually output from :func:`glmFit`
    coef : array_like of integers or strings
        vector indicating which coefficients of the linear model are to be
        tested equal to zero. Values must be column indices or column names of
        :code:`design`. Defaults to the last coefficient. Ignored if
        :code:`contrast` is specified.
    contrast : array or matrix of integers
        vector or matrix specifying one or more contrasts of the linear model
        coefficients to be tested equal to zero. Number of rows must equal to
        the number of columns of :code:`design`. If specified, then takes
        precedence over :code:`coef`.

    Returns
    -------
    DGELRT
        dataframe with two additional components:

        - :code:`fit` containing the result of :func:`glmFit`
        - :code:`comparison`, string describing the coefficient or the contrast
          being tested

        The dataframe has the same rows as :code:`y` and is ready to be
        displayed by :func:`topTags`. It contains the following columns:

        - :code:`"log2FoldChange"`, log2-fold-change of expression between
          conditions being tested.
        - :code:`"lfcSE"`, standard error of log2-fold-change.
        - :code:`"logCPM"`, average log2-counts per million, the average taken
          over all libraries in :code:`y`.
        - :code:`"stat"`, likelihood ratio statistics.
        - :code:`"pvalue"`, *p*-values.
    """
    if coef is None:
        coef = glmfit.design.shape[1] - 1
    if not isinstance(glmfit, DGEGLM):
        raise ValueError("glmfit must be a DGEGLM object (usually produced by glmFit).")

    if glmfit.AveLogCPM is None:
        glmfit.AveLogCPM = glmfit.aveLogCPM()
    nlibs = glmfit.coefficients.shape[1]

    # check design matrix
    design = glmfit.design
    nbeta = design.shape[1]
    if nbeta < 2:
        raise ValueError(
            "Need at least two columns for design, usually the first is the intercept columns"
        )
    coef_names = np.array(design.design_info.column_names)

    # Evaluate logFC for coef to be tested
    # Note that contrast takes precedence over coef: if contrast is given then reform
    # design matrix so that contrast of interest is last column
    if contrast is None:
        if not isinstance(coef, (list, np.ndarray)):
            coef = [coef]
        if isinstance(coef[0], str):
            check_coef = np.isin(coef, design.design_info.column_names)
            if (~check_coef).any():
                raise ValueError(
                    "One or more named coef arguments do not match a column of the design matrix."
                )
            coef_name = coef
            coef = np.nonzero([design.design_info.column_names == c for c in coef])[0]
        else:
            coef_name = [coef_names[c] for c in coef]
        logFC = glmfit.coefficients[:, coef] / np.log(2)
        lfcSE = glmfit.coeff_SE[:, coef] / np.log(2)
    else:
        contrast = np.array(contrast)
        if contrast.ndim > 2:
            raise ValueError("contrast must be 1-D or 2-D")
        if contrast.ndim < 2:
            contrast = contrast.reshape(contrast.shape[0], 1)
        if contrast.shape[0] != glmfit.coefficients.shape[1]:
            raise ValueError(
                "contrast vector of wrong length, should be equal to number of coefficients in the linear model"
            )
        ncontrasts = np.linalg.matrix_rank(contrast)
        Q, R = np.linalg.qr(contrast, mode="complete")
        if ncontrasts == 0:
            raise ValueError("contrasts are all zero")
        coef = np.arange(ncontrasts)
        logFC = (glmfit.coefficients @ contrast) / np.log(2)
        lfcSE = (glmfit.coeff_SE @ contrast) / np.log(2)
        if ncontrasts > 1:
            coef_name = f"LR test on {ncontrasts} degrees of freedom"
        else:
            contrast = np.squeeze(contrast)
            i = contrast != 0
            coef_name = " ".join(
                [f"{a}*{b}" for a, b in zip(contrast[i], coef_names[i])]
            )
        Dvec = np.ones(nlibs, int)
        Dvec[coef] = np.diag(R)[coef]
        Q = Q * Dvec
        design = design @ Q

    # Null design matrix
    non_coef = np.setdiff1d(np.arange(design.shape[1]), coef)
    design0 = design[:, non_coef]

    # Null fit
    fit_null = glmFit(
        glmfit.counts,
        design=design0,
        offset=glmfit.offset,
        weights=glmfit.weights,
        dispersion=glmfit.dispersion,
        prior_count=0,
    )

    # Likelihood ratio statistic
    LR = np.subtract(fit_null.deviance, glmfit.deviance)
    df_test = fit_null.df_residual - glmfit.df_residual
    LRT_pvalue = scipy.stats.chi2.sf(LR, df=df_test)
    tab = pd.DataFrame()
    if logFC.ndim > 1:
        for i in range(logFC.shape[1]):
            tab[f"logFC{i}"] = logFC[:, i]
            tab[f"lfcSE{i}"] = lfcSE[:, i]
        tab.columns = [
            "log2FoldChange" if i % 2 == 0 else "lfcSE"
            for i in range(2 * logFC.shape[1])
        ]

    else:
        tab["log2FoldChange"] = logFC
        tab["lfcSE"] = lfcSE
    tab["logCPM"] = glmfit.AveLogCPM
    tab["stat"] = LR
    tab["pvalue"] = LRT_pvalue
    tab.index = glmfit.counts.index
    res = DGELRT(tab, glmfit)
    res.comparison = coef_name
    res.df_test = df_test
    return res



from .addPriorCount import addPriorCount as addPriorCount
from .adjustedProfileLik import adjustedProfileLik as adjustedProfileLik
from .aveLogCPM import aveLogCPM as aveLogCPM
from .binomTest import binomTest as binomTest
from .DGEGLM import DGEGLM as DGEGLM
from .DGEList import DGEList as DGEList
from .dispCoxReid import dispCoxReid as dispCoxReid
from .dispCoxReidInterpolateTagwise import (
    dispCoxReidInterpolateTagwise as dispCoxReidInterpolateTagwise,
)
from .estimateGLMCommonDisp import estimateGLMCommonDisp as estimateGLMCommonDisp
from .estimateGLMTagwiseDisp import estimateGLMTagwiseDisp as estimateGLMTagwiseDisp
from .exactTest import exactTest as exactTest
from .exactTestBetaApprox import exactTestBetaApprox as exactTestBetaApprox
from .exactTestByDeviance import exactTestByDeviance as exactTestByDeviance
from .exactTestBySmallP import exactTestBySmallP as exactTestBySmallP
from .exactTestDoubleTail import exactTestDoubleTail as exactTestDoubleTail
from .glmFit import glmFit as glmFit
from .glmFit import glmLRT as glmLRT
from .glmQLFit import glmQLFit as glmQLFit
from .glmQLFit import glmQLFTest as glmQLFTest
from .glmQLFit import plotQLDisp as plotQLDisp
from .maximizeInterpolant import maximizeInterpolant as maximizeInterpolant
from .mglmLevenberg import mglmLevenberg as mglmLevenberg
from .mglmOneGroup import mglmOneGroup as mglmOneGroup
from .mglmOneWay import designAsFactor as designAsFactor
from .mglmOneWay import mglmOneWay as mglmOneWay
from .movingAverageByCol import movingAverageByCol as movingAverageByCol
from .nbinomDeviance import nbinomDeviance as nbinomDeviance
from .predFC import predFC as predFC
from .q2qnbinom import q2qnbinom as q2qnbinom
from .splitIntoGroups import splitIntoGroups as splitIntoGroups
from .stats import pnbinom as pnbinom
from .stats import qnbinom as qnbinom
from .systematicSubset import systematicSubset as systematicSubset
from .topTags import topTags as topTags
from .validDGEList import validDGEList as validDGEList

from .edgepy_cpp import *  # noqa: F403



# -----------------------------------------------------------------------------
# Copyright (C) 2004-2022 Gordon Smyth, Yifang Hu, Matthew Ritchie, Jeremy Silver, James Wettenhall, Davis McCarthy, Di Wu, Wei Shi, Belinda Phipson, Aaron Lun, Natalie Thorne, Alicia Oshlack, Carolyn de Graaf, Yunshun Chen, Mette Langaas, Egil Ferkingstad, Marcus Davy, Francois Pepin, Dongseok Choi
# Copyright (C) 2024 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# This file is based on the file 'R/decidetests.R' of the Bioconductor limma package (version 3.55.1).

import numpy as np
import pandas as pd
import scipy

from ..utils import cov2cor


class TestResults(pd.DataFrame):
    """
    A matrix-based class for storing the results of simultaneous tests.

    Instances of this class are usually created by :func:`decideTests`.

    A :class:`TestResults` object is essentially a matrix with elements equal
    to :math:`0`, :math:`1` or :math:`-1`. Zero represents acceptance of the
    null hypothesis, :math:`1` indicates rejection in favor of the right tail
    alternative and :math:`-1` indicates rejection in favor of the left tail
    alternative.

    :class:`TestResults` objects can be created by :code:`TestResults(results)`
    where :code:`results` is a matrix.
    """

    _metadata = [
        "levels",
        "labels",
    ]

    @property
    def _constructor(self):
        return TestResults

    @property
    def _constructor_sliced(self):
        return pd.Series

    def __init__(self, df, *args, **kwargs):
        super().__init__(df, *args, **kwargs)
        self.levels = None
        self.labels = None

    def summary(self):
        """
        Summary of the test results

        Returns
        -------
        pd.DataFrame
            a summary table
        """
        levels = self.levels
        if levels is None:
            levels = [-1, 0, 1]
        nlevels = len(levels)
        tab = pd.DataFrame(np.zeros((nlevels, self.shape[1]), dtype=int))
        labels = self.labels
        if labels is None:
            labels = levels
        tab.index = labels
        tab.columns = self.columns
        for i in range(nlevels):
            tab.iloc[i, :] = np.nan_sum(self == levels[i], axis=1)
        return tab


class FStat(np.ndarray):
    """
    Helper class to store F-statistics in a numpy array along with two degrees of freedom attributes

    Subclassing ndarray is necessarily to add attributes to it.

    Attributes
    ----------
    df1 : int
        first degrees-of-freedom parameter of the F-statistics
    df2 : int
        second degrees-of-freedom parameter of the F-statistics
    """

    def __new__(cls, input_array, df1=None, df2=None):
        obj = np.asarray(input_array).view(cls)
        obj.df1 = df1
        obj.df2 = df2
        return obj

    def __array_finalize__(self, obj):
        if obj is None:
            return
        self.df1 = getattr(obj, "df1", None)
        self.df2 = getattr(obj, "df2", None)


def classifyTestsF(self, cor_matrix=None, df=np.inf, p_value=0.01, fstat_only=False):
    """
    For each gene, classify a series of related *t*-statistics as significantly up or down using nested *F*-tests.

    This function implements the :code:`nestedF` multiple testing option
    offered by :func:`decideTests`. Users should generally use
    :func:`decideTests` rather than calling :func:`classifyTestsF` directly
    because, by itself, :func:`classifyTestsF` does not incorporate any
    multiple testing adjustment across genes. Instead, it simply tests across
    contrasts for each gene individually.

    :func:`classifyTestsF` used a nested *F*-test approach giving particular
    attention to correctly classifying genes that have two or more significant
    *t*-statistics, *i.e.* which are differentially expressed in two or more
    conditions.  For each row of :code:`tstat`, the overall *F*-statistics is
    constructed from the *t*-statistics as for :code:`FStat`.
    At least one contrast will be classified as significant if and only if the
    overall *F*-statistic is significant.
    If the overall *F*-statistic is significant, then the function makes a best
    choice as to which *t*-statistics contributed to this result.
    The methodology is based on the principle that any *t*-statistic should be
    called significant if the *F*-test is still significant for that row when
    all the larger *t*-statistics are set to the same absolute size as the
    *t*-statistic in question.

    Compared to conventional multiple testing methods, the nested *F*-test
    approach achieves better consistency between related contrasts. (For
    example, if B is judged to be different from C, then at least one of B or C
    should be different to A.) the approach was first used by [Michaud2008]_.
    The nested *F*-test approach provides *weak* control of the family-wise
    error rate, *i.e.* it correctly controls the type I error rate of calling
    any contrast as significant if all the null hypotheses are true. In other
    words, it provides error rate control at the overall *F*-test level but
    does not provide strict error rate control at the individual contrast
    level.

    Usually, :code:`self` is a limma linear model fitted object, from which a
    matrix of *t*-statistics can be extracted, but it can also be a numeric
    matrix of *t*-statistics. In either case, rows correspond to genes and
    columns to coefficients or contrasts. The :code:`cor_matrix` is the same as
    the correlation matrix of the coefficients from which the *t*-statistics
    were calculated and :code:`df` is the degrees of freedom of the
    *t*-statistics. All statistics for the same gene must have the same degrees
    of freedom.

    If :code:`fstat_only=True`, this function just returns the vector of
    overall *F*-statistics for each gene.

    Arguments
    ---------
    self : MArrayLM or ndarray
        matrix of *t*-statistics, or a :class:`MArrayLM` object from which the
        *t*-statistics may be extracted
    cor_matrix : ndarray
        covariance matrix of each of *t*-statistics. Will be extracted
        automatically from the :class:`MArrayLM` object, but otherwise defaults
        to the identity matrix.
    df : array_like
        array of degrees of freedom for the *t*-statistics. Should be
        broadcastable to the shape of :code:`tstats`. Will be extract
        automatically from the :class:`MArrayLM` object but otherwise defaults
        to :code:`np.inf`.
    p_value : float
        value between 0 and 1 giving the desired size of the test
    fstat_only : bool
        if :code:`True` then return the overall *F*-statistic as for
        :code:`FStat` instead of classifying the test results.

    Returns
    -------
    TestResults or ndarray
        if :code:`fstats_only=False`, then an object of class
        :class:`TestResults`, which is essentially a matrix with elements -1, 0
        or 1 depending on whether each *t*-statistics is classified as
        significantly negative, not significant or significantly positive
        respectively.
        if :code:`fstats_only=True`, then an array of *F*-statistics is
        returned with attributes :code:`df1` and :code:`df2` giving the
        corresponding degrees of freedom.
    """
    if isinstance(self, np.ndarray):
        tstat = self.copy()
    else:
        if self.t is None:
            raise ValueError("tstat cannot be extracted from object")
        if (cor_matrix is None) and (self.cov_coefficients is not None):
            # Check for an adjust any coefficient variances exactly zero (usually
            # caused by an all zero contrast)
            n = self.cov_coefficients.shape[0]
            if isinstance(self.cov_coefficients, pd.DataFrame):
                tmp = self.cov_coefficients.values
            else:
                tmp = self.cov_coefficients
            for i in range(n):
                if tmp[i, i] == 0:
                    tmp[i, i] = 1
            cor_matrix = cov2cor(self.cov_coefficients)
        if (
            (df is None)
            and (self.df_prior is not None)
            and (self.df_residual is not None)
        ):
            df = self.df_prior + self.df_residual
        tstat = self.t.copy()

    ngenes, ntests = tstat.shape
    if ntests == 1:
        if fstat_only:
            fstat = FStat(tstat**2)
            fstat.df1 = 1
            fstat.df2 = df
            return fstat
        else:
            p = 2 * scipy.stats.t.sf(np.abs(tstat), df)
            return TestResults(np.sign(tstat) * (p < p_value))

    # cor_matrix is estimated correlation matrix of the coefficients and also
    # the estimated covariance matrix of the t-statistics
    if cor_matrix is None:
        r = ntests
        Q = np.eye(r) / np.sqrt(r)
    else:
        evalues, evectors = np.linalg.eigh(cor_matrix)
        r = np.sum((evalues / evalues[0]) > 1e-8)
        Q = evectors[:, :r] * (1.0 / np.sqrt(evalues[:r]) / np.sqrt(r))

    # return overall moderated F-statistic only
    if fstat_only:
        fstat = FStat(np.squeeze((tstat @ Q) ** 2 @ np.ones((r, 1))))
        fstat.df1 = r
        fstat.df2 = df
        return fstat

    # return TestResults matrix
    if df > 1e10:
        # if qF ~ F(r, df), then r qF ~ chi2(r) when df -> infinity
        qF = scipy.stats.chi2.isf(p_value, r) / r
    else:
        qF = scipy.stats.f.isf(p_value, dfn=r, dfd=df)
    qF = np.broadcast_to(qF, ngenes)
    result = np.zeros((ngenes, ntests), dtype=int)
    for i in range(ngenes):
        x = tstat[i, :]
        if np.isnan(x).any():
            result[i, :] = np.nan
        else:
            if ((Q.T @ x).T @ (Q.T @ x)) > qF[i]:
                ord = np.flip(np.argsort(np.abs(x)))
                result[i, ord[0]] = np.sign(x[ord[0]])
                for j in range(1, ntests):
                    bigger = ord[:j]
                    x[bigger] = np.sign(x[bigger]) * np.abs(x[ord[j]])
                    if ((Q.T @ x).T @ (Q.T @ x)) > qF[i]:
                        result[i, ord[j]] = np.sign(x[ord[j]])
                    else:
                        break
    return TestResults(result)



import unittest

import numpy as np

from inmoose.limma import uniquegenelist, unwrapdups


class Test(unittest.TestCase):
    def test_unwrapdups(self):
        M = np.arange(12).reshape((4, 3))
        self.assertTrue(np.array_equal(unwrapdups(M, ndups=1), M))
        self.assertTrue(np.array_equal(unwrapdups(M, ndups=1, spacing=2), M))
        self.assertTrue(
            np.array_equal(unwrapdups(M), [[0, 3, 1, 4, 2, 5], [6, 9, 7, 10, 8, 11]])
        )
        self.assertTrue(
            np.array_equal(
                unwrapdups(M, ndups=4), [[0, 3, 6, 9, 1, 4, 7, 10, 2, 5, 8, 11]]
            )
        )
        self.assertTrue(
            np.array_equal(
                unwrapdups(M, ndups=2, spacing=2),
                [[0, 6, 1, 7, 2, 8], [3, 9, 4, 10, 5, 11]],
            )
        )

    def test_uniquegenelist(self):
        M = np.arange(12).reshape((4, 3))
        self.assertTrue(np.array_equal(uniquegenelist(M, ndups=1), M))
        self.assertTrue(np.array_equal(uniquegenelist(M, ndups=1, spacing=2), M))
        self.assertTrue(np.array_equal(uniquegenelist(M), [[0, 1, 2], [6, 7, 8]]))
        self.assertTrue(np.array_equal(uniquegenelist(M, ndups=4), [[0, 1, 2]]))
        self.assertTrue(
            np.array_equal(
                uniquegenelist(M, ndups=2, spacing=2), [[0, 1, 2], [3, 4, 5]]
            )
        )

        M = np.arange(6)
        self.assertTrue(np.array_equal(uniquegenelist(M, ndups=1), M))
        self.assertTrue(np.array_equal(uniquegenelist(M, ndups=1, spacing=2), M))
        self.assertTrue(np.array_equal(uniquegenelist(M), [0, 2, 4]))
        self.assertTrue(np.array_equal(uniquegenelist(M, ndups=3), [0, 3]))
        self.assertTrue(np.array_equal(uniquegenelist(M, ndups=3, spacing=2), [0, 1]))
        self.assertTrue(
            np.array_equal(uniquegenelist(M, ndups=2, spacing=3), [0, 1, 2])
        )



# distutils: language = c++
#-----------------------------------------------------------------------------
# Copyright (C) 2008-2022 Yunshun Chen, Aaron TL Lun, Davis J McCarthy, Matthew E Ritchie, Belinda Phipson, Yifang Hu, Xiaobei Zhou, Mark D Robinson, Gordon K Smyth
# Copyright (C) 2022-2025 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#-----------------------------------------------------------------------------

# This file contains Cython ports of the original C++ code from various files of
# the Bioconductor edgeR package (version 3.38.4):
# - 'src/nbdev.cpp' and 'src/R_compute_nbdev.cpp' (function
#   `compute_unit_nb_deviance`)
# - 'src/R_compute_apl.cpp' and 'src/adj_coxreid.cpp' (functions `compute_apl`
#   and `acr_compute`)
# - 'src/glm_one_group.cpp' (function `glm_one_group_cython`)
# - 'R/q2qnbinom.R' (function _q2qnbinom)
# - 'src/interpolator.cpp' and 'src/interpolator.h' (class `interpolator` and function
#   `quad_solver`)
# - 'src/R_maximize_interpolant.cpp' (function `maximize_interpolant` function)

import numpy as np
cimport cython
from libcpp.cmath cimport abs, log, isfinite, exp, isnan
from libc.math cimport INFINITY
from libc.stdint cimport int64_t

from libc.math cimport sqrt
from scipy.special cimport cython_special as sp
from scipy.special.cython_special cimport gammaln as lgamma
from scipy.linalg.lapack import get_lapack_funcs


cdef public ndarray[double, ndim=1] vector2ndarray "vector2ndarray"(const vector.vector[double]& data):
    return np.asarray(data)


ctypedef fused count_type:
    int64_t
    double


cdef double low_value = 1e-10
cdef double log_low_value = log(low_value)
cdef double mildly_low_value = 1e-8


@cython.ufunc
@cython.cdivision(True)
cdef double compute_unit_nb_deviance(double y, double mu, double phi):
    """
    Calculate the deviance of a negative binomial fit

    Note the protection for very large mu*phi (where we use a Gamma instead) or
    very small mu*phi (where we use a Poisson instead). This approximation
    protects against numerical instability introduced by subtracting a very
    large log value in (log mu) with another very large logarithm (log mu+1/phi).
    We need to consider the phi as the approximation is only good when the
    product is very big or very small.

    Arguments
    ---------
    y : array_like
        counts matrix
    mu : array_like
        expected means matrix (broadcastable to the shape of :code:`y`)
    phi : array_like
        dispersion matrix (broadcastable to the shape of :code:`y`)

    Returns
    -------
    ndarray
        matrix of deviances (same shape as :code:`y`)
    """
    # add a small value to protect against zero during division and log
    y = y + mildly_low_value
    mu = mu + mildly_low_value

    # Calculating the deviance using either the Poisson (small phi*mu), the
    # Gamma (large phi*mu) or NB (everything else). Some additional work is
    # put in to make the transitiosn between families smooth
    resid = y - mu
    product = mu * phi

    if phi < 1e-4:
        return 2 * (y * log(y/mu) - resid - 0.5*resid*resid*phi*(1+phi*(2/3*resid-y)))
    elif product > 1e6:
        return 2 * (resid/mu - log(y/mu)) * mu/(1+product)
    else:
        return 2 * (y * log(y/mu) + (y + 1/phi) * log((mu + 1/phi)/(y + 1/phi)))

@cython.ufunc
cdef double _q2qnbinom(double x, double input_mean, double output_mean, double dispersion):
    """
    Interpolated quantile to quantile mapping between negative-binomial distributions with the same dispersion but different means.

    This is the low-level Cythonized function called by :func:`q2qnbinom`.

    See also
    --------
    q2qnbinom

    Arguments
    ---------
    x : array_like
        matrix of counts
    input_mean : array_like
        matrix of population means for :code:`x`. If 1-D, then of the same
        length as :code:`x.shape[0]`
    output_mean : array_like
        matrix of population means for the output values. If 1-D, then of the
        same length as :code:`x.shape[0]`
    dispersion : array_like
        scalar, vector or matrix giving negative binomial dispersion values

    Returns
    -------
    ndarray
        matrix of same dimensions as :code:`x`, with :code:`output_mean` as the
        new nominal population mean
    """
    if x < 0:
        raise ValueError("x must be non-negative")
    if input_mean < 0:
        raise ValueError("input_mean must be non-negative")
    if output_mean < 0:
        raise ValueError("output_mean must be non-negative")
    if dispersion < 0:
        raise ValueError("dispersion must be non-negative")

    eps = 1e-14
    if input_mean < eps or output_mean < eps:
        input_mean += 0.25
        output_mean += 0.25
    ri = 1 + dispersion * input_mean
    vi = input_mean * ri
    ro = 1 + dispersion * output_mean
    vo = output_mean * ro

    q1 = output_mean + sqrt(vo / vi) * (x - input_mean)

    if x >= input_mean:
        p2 = sp.gammaincc(input_mean / ri, x/ri)
        q2 = sp.gammainccinv(output_mean / ro, p2) * ro
    else:
        p2 = sp.gammainc(input_mean / ri, x/ri)
        q2 = sp.gammaincinv(output_mean / ro, p2) * ro

    return (q1+q2)/2


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cpdef ndarray compute_apl(const count_type[:,:] y, const double[:,:] means, const double[:,:] disps, const double[:,:] weights, bool adjust, ndarray design):
    """
    Compute adjusted profile log-likelihoods of genewise negative binomial GLMs

    This is the low-level function called by :func:`adjustedProfileLik`.

    Arguments
    ---------
    y : array_like
        matrix of counts
    means : array_like
        matrix of expected means, same shape as :code:`y`
    disps : array_like
        matrix of dispersions
    weights : array_like
        matrix of observation weights, same shape as :code:`y`
    adjust : bool
        whether to use Cox-Reid adjustment
    design : array_like
        the design matrix

    Returns
    -------
    ndarray
        the genewise APL (one element per row in :code:`y`)
    """

    cdef Py_ssize_t tag, lib
    cdef double curmu, cury, curd, r, logmur, adj
    cdef long lwork = -1

    if adjust:
        routine = get_lapack_funcs('sytrf_lwork', dtype=np.double)
        ret = routine(design.shape[1])
        assert len(ret) == 2
        if ret[1] != 0:
            raise ValueError("Internal work array size computation failed: "
                             "%d" % (ret[1],))
        lwork = int(ret[0]+0.5)
        if lwork < 1:
            lwork = 1

    res = np.zeros(y.shape[0])
    cdef double[:] sum_loglike = res

    working_weights_array = np.zeros(y.shape[1])
    cdef double[:] working_weights = working_weights_array

    for tag in range(y.shape[0]):
        for lib in range(y.shape[1]):
            # mean should only be zero if count is zero, where the
            # log-likelihood would then be 0.
            if means[tag,lib] == 0:
                continue

            # each y is assumed to be the average of 'weights' counts, so we
            # convert from averages to the "original sums" in order to compute
            # NB probabilities
            curmu = means[tag,lib] * weights[tag,lib]
            cury = y[tag,lib] * weights[tag,lib]
            curd = disps[tag,lib] / weights[tag,lib]

            # compute the log-likelihood
            r = 1 / curd
            logmur = log(curmu + r)

            if curd > 0:
                sum_loglike[tag] += cury*log(curmu) - cury*logmur + r*log(r) - r*logmur + lgamma(cury+r) - lgamma(cury+1) - lgamma(r)
            else:
                sum_loglike[tag] += cury*log(curmu) - curmu - lgamma(cury+1)

            # adding the Jacobian, to account for the fact that we actually
            # want the log-likelihood of the _scaled_ NB distribution (after
            # dividing the original sum by the weight).
            sum_loglike[tag] += log(weights[tag,lib])

            if adjust:
                # computing W, the matrix of NB working weights
                # this is used to compute the Cox-Reid adjustment factor
                working_weights[lib] = curmu / (1 + curd*curmu)

        if adjust:
            if design.shape[1] == 1:
                adj = working_weights_array.sum()
                adj = 0.5*log(abs(adj))
                sum_loglike[tag] -= adj
            else:
                sum_loglike[tag] -= acr_compute(working_weights, design, lwork)

    return res


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef double acr_compute(const double[:] wptr, ndarray design, long lwork):
    """
    Compute the Cox-Reid adjustment factor

    XtWX represents the expected Fisher information. The overall strategy is to
    compute the log-determinant of this matrix, to compute the adjustment factor
    for the likelihood (in order to account for uncertainty in the nuisance
    parameters i.e. the fitted values).

    We want to apply the Cholesky decomposition to the XtWX matrix. However, to
    be safe, we call the routine to do a symmetric indefinite factorization i.e.
    A = LDLt. This guarantees factorization for singular matrices when the
    actual Cholesky decomposition would fail because it would start whining
    about non-positive eigenvectors.

    We then try to compute the determinant of XtWX, using two facts:

    - for triangular matrices, the determinant is the product of the diagonals
    - det(LDL*) = det(L)*det(D)*det(L*)
    - all diagonal elements of L are unity

    Thus, all we need to do is to sum over all log'd diagonal elements in D. We
    then divide by two, because that is just the definition of the Cox-Reid
    adjustment.

    If one of the diagonal elements is zero or NA, we replace it with an
    appropriately small non-zero value. This is valid because the zero elements
    correspond to all-zero column in WX, which in turn only arises when there
    are fitted values of zero, which will be constant at all dispersions. Thus,
    any replacement value will eventually cancel out during interpolation to
    obtain the CRAPLE.

    Note that the scipy routine will also do some pivoting, essentially solving
    PAP* = LDL* for some permutation matrix P. This should not affect anything:
    the determinant of the permutation matrix is either 1 or -1, but it cancels
    out, so det(A) = det(PAP*).

    Further note that the routine can theoretically give block diagonals, but
    this should not occur for positive (semi)definite matrices, which is what
    XtWX should always be.

    Arguments
    ---------
    wptr : array_like
        weights array
    design : array_like
        the design matrix

    Returns
    -------
    float
        the Cox-Reid adjustment factor
    """
    cdef double res = 0
    cdef Py_ssize_t i

    xtwx = (design.T * wptr) @ design

    # BEGIN LDL* decomposition
    solver = get_lapack_funcs('sytrf', dtype=np.double)
    ldu, _, info = solver(xtwx, lwork=lwork, lower=True, overwrite_a=False)
    if info < 0:
        raise ValueError('{} exited with the internal error "illegal value '
                         'in argument number {}". See LAPACK documentation '
                         'for the error codes.'.format('DSYTRF', -info))
    # END LDL* decomposition

    cdef double[:,:] d = ldu

    # log-determinant as sum of the log-diagonals, then halving
    assert d.shape[0] == d.shape[1]
    for i in range(ldu.shape[0]):
        if d[i,i] < low_value or not isfinite(d[i,i]):
            res += log_low_value
        else:
            res += log(d[i,i])

    return 0.5*res


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cpdef (double, bool) glm_one_group_cython(Py_ssize_t nlibs,
                                          const count_type[:] counts, const double[:] offset,
                                          const double[:] disp, const double[:] weights,
                                          long maxit, double tolerance, double cur_beta):
    """
    Simplified fit for negative binomial GLM when the design matrix is one group

    This is the low-level function ultimately called by :func:`fit_one_group`.

    See Also
    --------
    glm_one_group
    fit_one_group

    Arguments
    ---------
    nlibs : Py_ssize_t
        number of libraries
    counts : array_like
        vector of counts for a given gene, of size `nlibs`
    offset : array_like
        vector of offsets for a given gene, of size `nlibs`
    disp : array_like
        vector of dispersions for a given gene, of size `nlibs`
    weights : array_like
        vector of observation weights for a given gene, of size `nlibs`
    maxit : long
        maximum number of Newton-Raphson iterations
    tol : float
        tolerance for convergence in the Newton-Raphson iteration
    cur_beta : float
        initial beta coefficient

    Returns
    -------
    float
        fitted coefficient
    bool
        whether the fit converged
    """

    cdef Py_ssize_t j
    cdef long i
    cdef double totweight, dl, info, mu, cur_val

    # Setting up initial values for beta as the log of the mean of the ratio of
    # counts to offsets. This is the exact solution for the gamma distribution
    # (which is the limit of the NB as the dispersion goes to infinity).
    # However, if cur_beta is not NA, then we assume it is good.
    nonzero = False
    if isnan(cur_beta):
        cur_beta = 0
        totweight = 0
        for j in range(nlibs):
            cur_val = counts[j]
            if cur_val > low_value:
                cur_beta += cur_val / exp(offset[j]) * weights[j]
                nonzero = True
            totweight += weights[j]
        cur_beta = log(cur_beta / totweight)
    else:
        for j in range(nlibs):
            if counts[j] > low_value:
                nonzero = True
                break

    # skipping to a result for all-zero rows
    if not nonzero:
        return (-INFINITY, True)

    # Newton-Raphson iteration to converge to mean
    has_converged = False
    for i in range(maxit):
        dl = 0
        info = 0

        for j in range(nlibs):
            mu = exp(cur_beta + offset[j])
            denom = 1 + mu*disp[j]
            dl += (counts[j] - mu) / denom * weights[j]
            info += mu / denom * weights[j]
        step = dl / info
        cur_beta += step
        if abs(step) < tolerance:
            has_converged = True
            break

    return (cur_beta, has_converged)


cdef struct solution:
    double sol1, sol2
    bool solvable

cdef solution quad_solver(double a, double b, double c):
    """
    Solve quadratic equation ax^2 + bx + c = 0.

    Parameters
    ----------
    a : double
        Coefficient of x^2
    b : double
        Coefficient of x
    c : double
        Constant term

    Returns
    -------
    solution
        Structure containing the solutions and whether they exist
    """
    cdef:
        solution cur_sol
        double back, front

    if a == 0 and b == 0:
        if c == 0:
            cur_sol.solvable = True
        else:
            cur_sol.solvable = False
        return cur_sol

    if a == 0:
        cur_sol.sol1 = -c/b
        cur_sol.sol2 = -c/b
        cur_sol.solvable = True
        return cur_sol

    back = b * b - 4 * a * c
    if back < 0:
        cur_sol.solvable = False
        return cur_sol

    front = -b / (2 * a)
    back = sqrt(back) / (2 * a)
    cur_sol.sol1 = front - back
    cur_sol.sol2 = front + back
    cur_sol.solvable = True
    return cur_sol


cdef class interpolator:
    """
    Class to identify the global maximum in the interpolating function.

    This is a Cython port of the C++ interpolator class from edgeR.

    Attributes
    ----------
    npts : Py_ssize_t
        number of points in cubic spline fit
    b : vector[double]
        vector of coefficients of degree 1 for the cubic spline
    c : vector[double]
        vector of coefficients of degree 2 for the cubic spline
    d : vector[double]
        vector of coefficients of degree 3 for the cubic spline
    """

    cdef:
        Py_ssize_t npts
        vector.vector[double] b, c, d

    def __init__(self, Py_ssize_t n):
        """
        Initialize the interpolator with the number of points.

        Parameters
        ----------
        n : size_t
            Number of points for interpolation (must be at least 2)
        """
        if n < 2:
            raise RuntimeError("must have at least two points for interpolation")
        self.npts = n
        self.b.resize(n)
        self.c.resize(n)
        self.d.resize(n)

    cdef double find_max(self, const double* x, const double* y):
        """
        Find the maximum of the interpolating spline.

        Parameters
        ----------
        x : const double*
            Array of x-coordinates
        y : const double*
            Array of y-coordinates

        Returns
        -------
        double
            The x-coordinate where the maximum occurs
        """
        cdef:
            double maxed = -1
            Py_ssize_t maxed_at = -1
            Py_ssize_t i
            double x_max

            double ld, lc, lb
            double rd, rc, rb
            solution sol_left, sol_right
            double chosen_sol, temp

        # Get initial guess for MLE
        for i in range(self.npts):
            if maxed_at == -1 or y[i] > maxed:
                maxed = y[i]
                maxed_at = i

        x_max = x[maxed_at]

        # Fit the spline
        self._fmm_spline(x, y)

        # Check left segment for maximum
        if maxed_at > 0:
            ld = self.d[maxed_at-1]
            lc = self.c[maxed_at-1]
            lb = self.b[maxed_at-1]

            sol_left = quad_solver(3*ld, 2*lc, lb)
            if sol_left.solvable:
                # Use solution with maximum (not minimum)
                # If the curve is mostly increasing, the maximal point is located at the smaller solution (i.e. sol1 for a>0).
                # If the curve is mostly decreasing, the maximal point is located at the larger solution (i.e. sol1 for a<0).
                chosen_sol = sol_left.sol1

                # Check if solution is within segment bounds
                # The spline coefficients are designed such that 'x' in 'y + b*x + c*x^2 + d*x^3' is
                # equal to 'x_t - x_l' where 'x_l' is the left limit of that spline segment and 'x_t'
                # is where you want to get an interpolated value. This is necessary in 'splinefun' to
                # ensure that you get 'y' (i.e. the original data point) when 'x=0'. For our purposes,
                # the actual MLE corresponds to 'x_t' and is equal to 'solution + x_0'.
                if chosen_sol > 0 and chosen_sol < x[maxed_at] - x[maxed_at-1]:
                    temp = ((ld*chosen_sol + lc)*chosen_sol + lb)*chosen_sol + y[maxed_at-1]
                    if temp > maxed:
                        maxed = temp
                        x_max = chosen_sol + x[maxed_at-1]

        # Check right segment for maximum
        if maxed_at < self.npts - 1:
            rd = self.d[maxed_at]
            rc = self.c[maxed_at]
            rb = self.b[maxed_at]

            sol_right = quad_solver(3*rd, 2*rc, rb)
            if sol_right.solvable:
                chosen_sol = sol_right.sol1

                if chosen_sol > 0 and chosen_sol < x[maxed_at+1] - x[maxed_at]:
                    temp = ((rd*chosen_sol + rc)*chosen_sol + rb)*chosen_sol + y[maxed_at]
                    if temp > maxed:
                        maxed = temp
                        x_max = chosen_sol + x[maxed_at]

        return x_max

    cdef void _fmm_spline(self, const double* x, const double* y):
        """
        Fit a cubic spline using Forsythe Malcolm Moler algorithm.

        In this case the end-conditions are determined by fitting cubic
        polynomials to the first and last 4 points and matching the third
        derivatives of the spline at the end-points to the third derivatives of
        these cubics at the end-points.

        This function is a direct port from splines.c in the R stats package.
        https://github.com/SurajGupta/r-source/blob/master/src/library/stats/src/splines.c

        Parameters
        ----------
        x : const double*
            Array of x-coordinates
        y : const double*
            Array of y-coordinates
        """
        cdef:
            Py_ssize_t nm1, i
            double t

        if self.npts < 2:
            raise ValueError("spline interpolation requires at least 2 points")

        if self.npts < 3:
            t = (y[1] - y[0])
            self.b[0] = t / (x[1] - x[0])
            self.b[1] = self.b[0]
            self.c[0] = self.c[1] = self.d[0] = self.d[1] = 0.0
            return

        nm1 = self.npts - 1

        # Set up tridiagonal system
        # b = diagonal, d = offdiagonal, c = right hand side
        self.d[0] = x[1] - x[0]
        self.c[1] = (y[1] - y[0]) / self.d[0]

        for i in range(1, self.npts-1):
            self.d[i] = x[i+1] - x[i]
            self.b[i] = 2.0 * (self.d[i-1] + self.d[i])
            self.c[i+1] = (y[i+1] - y[i]) / self.d[i]
            self.c[i] = self.c[i+1] - self.c[i]

        # End conditions
        # Third derivatives at x[0] and x[n-1] obtained from divided differences

        self.b[0] = -self.d[0]
        self.b[self.npts-1] = -self.d[nm1-1]
        self.c[0] = self.c[self.npts-1] = 0.0

        if self.npts > 3:
            self.c[0] = self.c[2] / (x[3] - x[1]) - self.c[1] / (x[2] - x[0])
            self.c[self.npts-1] = self.c[nm1-1] / (x[self.npts-1] - x[self.npts-3]) - self.c[self.npts-3] / (x[nm1-1] - x[self.npts-4])
            self.c[0] = self.c[0] * self.d[0] * self.d[0] / (x[3] - x[0])
            self.c[self.npts-1] = -self.c[self.npts-1] * self.d[nm1-1] * self.d[nm1-1] / (x[self.npts-1] - x[self.npts-4])

        # Gaussian elimination
        for i in range(1, self.npts):
            t = self.d[i-1] / self.b[i-1]
            self.b[i] = self.b[i] - t * self.d[i-1]
            self.c[i] = self.c[i] - t * self.c[i-1]

        # Backward substitution
        self.c[self.npts-1] = self.c[self.npts-1] / self.b[self.npts-1]
        for i in range(nm1-1, -1, -1):
            self.c[i] = (self.c[i] - self.d[i] * self.c[i+1]) / self.b[i]

        # c[i] is now the sigma[i-1] of the text
        # Compute polynomial coefficients
        self.b[self.npts-1] = (y[self.npts-1] - y[self.npts-2]) / self.d[self.npts-2] + self.d[self.npts-2] * (self.c[self.npts-2] + 2.0 * self.c[self.npts-1])
        for i in range(nm1):
            self.b[i] = (y[i+1] - y[i]) / self.d[i] - self.d[i] * (self.c[i+1] + 2.0 * self.c[i])
            self.d[i] = (self.c[i+1] - self.c[i]) / self.d[i]
            self.c[i] = 3.0 * self.c[i]

        self.c[self.npts-1] = 3.0 * self.c[self.npts-1]
        self.d[self.npts-1] = self.d[nm1-1]



@cython.wraparound(False)
@cython.boundscheck(False)
cpdef vector.vector[double] maximize_interpolant(vector.vector[double] spts, const double[:,:] likelihoods):
    """
    Find the maximum of interpolating splines for each row in the likelihood matrix.

    This is a Cython port of the maximize_interpolant function from edgeR.

    Parameters
    ----------
    spts : vector[double]
        Spline points (x-coordinates)
    likelihoods : const double[:,:]
        Matrix of likelihood values, where each row represents a different tag/gene
        and each column corresponds to a spline point

    Returns
    -------
    vector[double]
        Vector of x-coordinates where the maximum occurs for each tag/gene

    Raises
    ------
    RuntimeError
        If the number of columns in likelihood matrix doesn't match the number of spline points
    """
    cdef:
        Py_ssize_t num_pts = spts.size()
        Py_ssize_t num_tags = likelihoods.shape[0]
        Py_ssize_t tag, i
        interpolator maxinterpol
        vector.vector[double] current_ll
        vector.vector[double] all_spts
        vector.vector[double] output

    if num_pts != likelihoods.shape[1]:
        raise RuntimeError("number of columns in likelihood matrix should be equal to number of spline points")

    maxinterpol = interpolator(num_pts)
    current_ll.resize(num_pts)
    all_spts = spts  # Make a copy to guarantee contiguousness
    output.resize(num_tags)

    for tag in range(num_tags):
        # Copy current row to working array
        for i in range(num_pts):
            current_ll[i] = likelihoods[tag, i]

        output[tag] = maxinterpol.find_max(&all_spts[0], &current_ll[0])

    return output



import unittest

import numpy as np
from patsy import dmatrix

from inmoose.limma import makeContrasts


class Test(unittest.TestCase):
    def test_makeContrasts(self):
        self.assertTrue(
            np.array_equal(
                makeContrasts(contrasts=["B-A", "C-B", "C-A"], levels=["A", "B", "C"]),
                [[-1, 0, -1], [1, -1, 0], [0, 1, 1]],
            )
        )
        self.assertTrue(
            np.array_equal(
                makeContrasts(contrasts="A-(B+C)/2", levels=["A", "B", "C"]),
                [[1.0], [-0.5], [-0.5]],
            )
        )

        design = dmatrix("0+A+B", {"A": ["a1", "a2", "a1"], "B": ["b1", "b2", "b3"]})
        self.assertTrue(
            np.array_equal(
                makeContrasts(contrasts="A[a1]-A[a2]", levels=design),
                [[1.0], [-1.0], [0.0], [0.0]],
            )
        )

        # test support for underscores in contrasts
        self.assertTrue(
            np.array_equal(
                makeContrasts(
                    contrasts=["T_T[B_1]-T[A_0]", "T[C_2]-T_T[B_1]", "T[C_2]-T[A_0]"],
                    levels=["T[A_0]", "T_T[B_1]", "T[C_2]"],
                ),
                [[-1, 0, -1], [0, 1, 1], [1, -1, 0]],
            )
        )



# -----------------------------------------------------------------------------
# Copyright (C) 2024-2025 L. Meunier, T. Tavernier

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

import base64
import os
from datetime import datetime
from io import BytesIO

import jinja2
import matplotlib.pyplot as plt

from .cohort_metric import CohortMetric


class QCReport:
    """
    Generate a cohort quality control report using Jinja2.
    """

    def __init__(self, cohort_qc: CohortMetric, template_dir: str = None):
        self.cohort_qc = cohort_qc

        if template_dir is None:
            template_dir = os.path.dirname(os.path.abspath(__file__))

        self.template_env = jinja2.Environment(
            loader=jinja2.FileSystemLoader(template_dir),
            autoescape=True,
            trim_blocks=True,
            lstrip_blocks=True,
        )
        self.template_env.filters["percentage"] = lambda v: f"{v * 100:.2f}%"
        self.template_env.filters["format_float"] = lambda v, p=4: f"{v:.{p}f}"

    @staticmethod
    def plot_to_base64(plot: plt.Figure) -> str:
        buffer = BytesIO()
        plot.savefig(buffer, format="png", bbox_inches="tight")
        buffer.seek(0)
        image_png = buffer.read()
        buffer.close()
        plt.close(plot)
        return base64.b64encode(image_png).decode("utf-8")

    def generate_report(self, template_name: str = "qc_report.html") -> str:
        covariates = [self.cohort_qc.batch_column] + self.cohort_qc.covariates

        pca_plots = []
        for covariate in covariates:
            before = None
            if self.cohort_qc.data_expression_df_before is not None:
                fig_before = self.cohort_qc.pca_plot(
                    self.cohort_qc.pcs_before,
                    self.cohort_qc.clinical_df[covariate],
                    title=f"PCA before correction - {covariate}",
                )
                before = self.plot_to_base64(fig_before)

            fig_after = self.cohort_qc.pca_plot(
                self.cohort_qc.pcs_after,
                self.cohort_qc.clinical_df[covariate],
                title=f"PCA after correction - {covariate}",
            )
            after = self.plot_to_base64(fig_after)

            pca_plots.append(
                {
                    "covariate": covariate,
                    "before": before,
                    "after": after,
                }
            )

        ylim = self.cohort_qc.pca_after.explained_variance_ratio_.max()
        if self.cohort_qc.data_expression_df_before is not None:
            ylim = max(ylim, self.cohort_qc.pca_before.explained_variance_ratio_.max())

        pca_variance_before = (
            self.plot_to_base64(
                self.cohort_qc.plot_pca_variance(
                    self.cohort_qc.pca_before.explained_variance_ratio_, ylim
                )
            )
            if self.cohort_qc.data_expression_df_before is not None
            else None
        )

        pca_variance_after = self.plot_to_base64(
            self.cohort_qc.plot_pca_variance(
                self.cohort_qc.pca_after.explained_variance_ratio_, ylim
            )
        )

        summary = self.cohort_qc.cohort_summary()
        samples_by_comb = summary.get("Samples by Covariate Combination", [])
        covariates = self.cohort_qc.covariates

        mad_before = self.cohort_qc.mad_before
        mad_after = self.cohort_qc.mad_after
        effect_metric = self.cohort_qc.effect_metric

        if len(self.cohort_qc.covariates) > 0:
            covariate_combinations = (
                self.cohort_qc.clinical_df.groupby(self.cohort_qc.covariates)
                .size()
                .index
            )
        else:
            covariate_combinations = [()]

        distribution_plots = []
        for combination in covariate_combinations:
            fig = self.cohort_qc.compare_sample_distribution_by_covariates(combination)
            img = self.plot_to_base64(fig)
            distribution_plots.append(
                {
                    "combination": combination,
                    "img": img,
                }
            )

        global_distribution_plot = self.plot_to_base64(
            self.cohort_qc.compare_sample_distribution_total()
        )

        silhouette_before = self.cohort_qc.silhouette_before
        silhouette_after = self.cohort_qc.silhouette_after

        entropy_before = self.cohort_qc.entropy_before
        entropy_after = self.cohort_qc.entropy_after

        summary_mixed = self.cohort_qc.summary
        mixed_proportions_table = [
            {**{f"cov{i + 1}": v for i, v in enumerate(comb)}, "proportion": proportion}
            for comb, proportion in summary_mixed[
                "overall_proportion_by_covariate_combination"
            ].items()
        ]

        mixed_dataset_ratio = summary_mixed["total_mixed_datasets"] / (
            summary_mixed["total_mixed_datasets"]
            + summary_mixed["total_non_mixed_datasets"]
        )
        mixed_sample_ratio = summary_mixed["total_mixed_samples"] / (
            summary_mixed["total_mixed_samples"]
            + summary_mixed["total_non_mixed_samples"]
        )

        def format_association_matrix(matrix):
            formatted = matrix.copy()
            for col in matrix.columns:
                formatted[col] = formatted[col].apply(
                    lambda v: f"<strong>statistics</strong>={v[0]}<br><strong>p-value</strong>={v[1]}<br><strong>test performed</strong>={v[2]}<br><strong>number of samples used</strong>={v[3]}"
                )
            return formatted.to_html(escape=False)

        context = {
            "datetime": datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
            "pca_plots": pca_plots,
            "pca_variance_before": pca_variance_before,
            "pca_variance_after": pca_variance_after,
            "association_matrix_before": format_association_matrix(
                self.cohort_qc.association_matrix_before
            ),
            "association_matrix_after": format_association_matrix(
                self.cohort_qc.association_matrix
            ),
            "summary": summary,
            "samples_by_comb": samples_by_comb,
            "covariates": covariates,
            "mad_before": mad_before,
            "mad_after": mad_after,
            "effect_metric": effect_metric,
            "distribution_plots": distribution_plots,
            "global_distribution_plot": global_distribution_plot,
            "silhouette_before": silhouette_before,
            "silhouette_after": silhouette_after,
            "entropy_before": entropy_before,
            "entropy_after": entropy_after,
            "summary_mixed": summary_mixed,
            "mixed_dataset_ratio": mixed_dataset_ratio,
            "mixed_sample_ratio": mixed_sample_ratio,
            "mixed_proportions_table": mixed_proportions_table,
            "covariate_names": self.cohort_qc.covariates,
        }

        template = self.template_env.get_template(template_name)
        return template.render(**context)

    def save_report(
        self,
        output_path: str = "cohort_qc_report.html",
        template_name: str = "qc_report.html",
    ):
        html = self.generate_report(template_name)
        with open(output_path, "w", encoding="utf-8") as f:
            f.write(html)



# -----------------------------------------------------------------------------
# Copyright (C) 2004-2022 Gordon Smyth, Yifang Hu, Matthew Ritchie, Jeremy Silver, James Wettenhall, Davis McCarthy, Di Wu, Wei Shi, Belinda Phipson, Aaron Lun, Natalie Thorne, Alicia Oshlack, Carolyn de Graaf, Yunshun Chen, Mette Langaas, Egil Ferkingstad, Marcus Davy, Francois Pepin, Dongseok Choi
# Copyright (C) 2024 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# This file is based on the file 'R/ebayes.R' of the Bioconductor limma package (version 3.55.1).


import logging

import numpy as np
import pandas as pd
import scipy

from .decidetests import classifyTestsF
from .marraylm import MArrayLM
from .squeezeVar import squeezeVar


def eBayes(
    fit,
    proportion=0.01,
    stdev_coef_lim=(0.1, 4),
    trend=False,
    robust=False,
    winsor_tail_p=(0.05, 0.1),
):
    """
    Empirical Bayes statistics for differential expression

    Given a linear model fit from :func:`lmFit`, compute moderated
    *t*-statistics, moderated *F*-statistics, and log-odds of differential
    expression by empirical Bayes moderation of the standard errors towards a
    global value.

    This function is used to rank genes in order of evidence for differential
    expression. They use an empirical Bayes method to squeeze the genewise
    residual variances towards a common value (or towards a global trend)
    [Smyth2004]_ [Phipson2016]_.  The degrees of freedom for the individual
    variances are increased to reflect the extra information gained from the
    empirical Bayes moderation, resulting in increased statistical power to
    detect differential expression.

    This function accepts as input a :class:`MArrayLM` fitted model :code:`fit`
    produced by :func:`lmFit`. The columns of :code:`fit` define a set of
    contrasts which are to be tested equal to zero. The fitted model object may
    have been processed by :func:`contrasts_fit` before being passed to
    :func:`eBayes` to convert the coefficients of the original design matrix
    into an arbitrary number of contrasts.

    The empirical Bayes moderated *t*-statistics test each individual contrast
    equal to zero. For each gene (row), the moderated *F*-statistic tests
    whether all the contrasts are zero. The *F*-statistic is an overall test
    computed from the set of *t*-statistics for that probe. This is exactly
    analogous to the relationship between *t*-tests and *F*-statistics in
    conventional ANOVA, except that the residual mean squares have been
    moderated between genes.

    The estimates :code:`s2_prior` and :code:`df_prior` are computed by
    :func:`fitFDist`. :code:`s2_post` is the weighted average of
    :code:`s2_prior` and :code:`sigma**2` with weights proportional to
    :code:`df_prior` and :code:`df_residual` respectively. The log-odds of
    differential expression :code:`lods` was called the *B-statistic* by
    [Loennstedt2002]_. The *F*-statistics :code:`F` are computed by
    :func:`classifyTestsF` with :code:`fstats_only=True`.

    :func:`eBayes` does not compute ordinary *t*-statistics because they always
    have worse performance than the moderated versions. The ordinary
    (unmoderated) *t*-statistics can, however, be easily extracted from the
    linear model output for comparison purposes -- see the example code below.

    The use of :func:`eBayes` with :code:`trend=True` is known as the
    *limma-trend* method [Law2014]_ [Phipson2016]_. With this option, an
    intensity-dependent trend is fitted to the prior variances
    :code:`s2_prior`. Specifically, :func:`squeezeVar` is called with the
    :code:`covariate` equal to :code:`Amean`, the average log2-intensity for
    each gene. The trend that is fitted can be examined by :func:`plotSA`.
    limma-trend is useful for processing expression values that show a
    mean-variance relationship. This is often useful for microarray data, and
    it can also be applied to RNA-Seq counts that have been converged to
    log2-counts per million (logCPM) values [Law2014]_. When applied to RNA-Seq
    logCPM values, limma-trend give similar results to the :func:`voom` method.
    The voom method incorporates the mean-variance trend into the precision
    weights, whereas limma-trend incorporates the trend into the empirical
    Bayed moderation. limma-trend is somewhat simpler than :code:`voom` because
    it assumes that the sequencing depths (library sizes) are wildly different
    between the samples and it applies the mean-variance trend on a genewise
    basis instead to individual observations. limma-trend is recommended for
    RNA-Seq analysis when the library sizes are reasonably consistent (less
    than 3-fold difference from smallest to largest) because of its simplicity
    and speed.

    If :code:`robust=True`, then the robust empirical Bayes procedure of
    [Phipson2016]_ is used. This is frequently useful to protect the empirical
    Bayes procedure against hyper-variable or hypo-variable genes, especially
    when analysing RNA-Seq data. See :func:`squeezeVar` for more details.

    Arguments
    ---------
    fit : MArrayLM
        fitted model object produce by :func:`lmFit` or :func:`contrasts_fit`
    proportion : float
        value between 0 and 1, assumed proportion of genes which are
        differentially expressed
    stdev_coef_lim : pair of floats
        assumed lower and upper limits for the standard deviations of
        log2-fold-changes for differentially expressed genes
    trend : bool or array_like
        whether an intensity-dependent trend should be allowed for the prior
        variance.  If :code:`False` then the prior variance is constant.
        Alternatively, :code:`trend` can be a row-wise array, which will be used
        as the covariate for the prior variance.
    robust : bool
        whether the estimation of :code:`df_prior` and :code:`var_prior` should
        be robustified against outlier sample variances.

    Returns
    -------
    MArrayLM
        an object containing everything found in :code:`fit` plus the following
        added attributes:

        - :code:`t`: matrix of moderated *t*-statistics
        - :code:`p_value`: matrix of two-sided *p*-values corresponding to the
          *t*-statistics
        - :code:`lods` matrix giving the log-odds of differential expression
          (on the natural log scale)
        - :code:`s2_prior`: estimated prior value for :code:`sigma**2`. A
          row-wise array if :code:`covariate` is non-:code:`None`, otherwise a
          single value
        - :code:`df_prior`: degrees of freedom associated with
          :code:`s2_prior`. A row-wise array if :code:`robust=True`, otherwise
          a single value
        - :code:`df_total`: row-wise array giving the total degrees of freedom
          associated with the *t*-statistics for each gene. Equal to
          :code:`df_prior + df_residual` or :code:`sum(df_residual)`, whichever
          is smaller.
        - :code:`s2_post`: row-wise array giving the posterior values for
          :code:`sigma**2`
        - :code:`var_prior`: column-wise array giving estimated prior values for
          the variance of the log2-fold-changes for differentially expressed
          gene for each contrast. Used for evaluating :code:`lods`.
        - :code:`F`: row-wise array of moderated *F*-statistics for testing all
          contrasts defined by the columns of :code:`fit` simultaneously equal
          to zero.
        - :code:`F_p_value`: row-wise array giving *p*-values corresponding to
          :code:`F`

        The matrices :code:`t`, :code:`p_value`, :code:`lods` have the same
        dimensions as the input object :code:`fit`, with rows corresponding to
        genes and columns to coefficients or constrats.  The vector
        :code:`s2_prior`, :code:`df_prior`, :code:`df_total`, :code:`F` and
        :code:`F_p_value` correspond to rows, with length equal to the number
        of genes. The vector :code:`var_prior` corresponds to columns, with
        length equal to the number of contrasts. If :code:`s2_prior` or
        :code:`df_prior` have length 1, then the same value applies to all
        genes.

        :code:`s2_prior`, :code:`df_prior` and :code:`var_prior` contain
        empirical Bayes hyperparameters used to obtain :code:`df_total`,
        :code:`s2_post` and :code:`lods`.
    """
    if not isinstance(fit, MArrayLM):
        raise ValueError("fit is not a valid MArrayLM object")
    if isinstance(trend, bool) and trend and (fit.Amean is None):
        raise ValueError("Need Amean component in fit to estimate trend")
    eb = _ebayes(
        fit=fit,
        proportion=proportion,
        stdev_coef_lim=stdev_coef_lim,
        trend=trend,
        robust=robust,
        winsor_tail_p=winsor_tail_p,
    )
    fit.df_prior = eb["df_prior"]
    fit.s2_prior = eb["s2_prior"]
    fit.var_prior = eb["var_prior"]
    fit.proportion = proportion
    fit.s2_post = eb["s2_post"]
    fit.t = eb["t"]
    fit.df_total = eb["df_total"]
    fit.p_value = eb["p_value"]
    fit.lods = eb["lods"]
    if (fit.design is not None) and np.linalg.matrix_rank(
        fit.design
    ) == fit.design.shape[1]:
        F_stat = classifyTestsF(fit, fstat_only=True)
        fit.F = np.asarray(F_stat)
        df1 = F_stat.df1
        df2 = F_stat.df2
        # if Y ~ F(df1, df2), then df1 Y ~ chi2(df1) when df2 -> infinity
        if df2 > 1e10:
            fit.F_p_value = scipy.stats.chi2.sf(df1 * fit.F, df=df1)
        else:
            fit.F_p_value = scipy.stats.f.sf(fit.F, dfn=df1, dfd=df2)

    return fit


def _ebayes(
    fit,
    proportion=0.01,
    stdev_coef_lim=(0.1, 4),
    trend=False,
    robust=False,
    winsor_tail_p=(0.05, 0.1),
):
    coefficients = fit.coefficients
    stdev_unscaled = fit.stdev_unscaled
    assert np.array_equal(coefficients.index, stdev_unscaled.index)
    sigma = fit.sigma
    df_residual = fit.df_residual
    if (
        (coefficients is None)
        or (stdev_unscaled is None)
        or (sigma is None)
        or (df_residual is None)
    ):
        raise ValueError("argument is not a valid lmFit object")
    if np.max(df_residual) == 0:
        raise ValueError("No residual degrees of freedom in linear model fits")
    if not np.any(np.isfinite(sigma)):
        raise ValueError("no finite residual standard deviations")
    if isinstance(trend, bool):
        if trend:
            covariate = fit.Amean
            if covariate is None:
                raise ValueError("need Amean component in fit to estimate trend")
        else:
            covariate = None
    else:
        covariate = np.broadcast_to(trend, sigma.shape)

    # Moderated t-statistic
    out = squeezeVar(
        sigma**2,
        df_residual,
        covariate=covariate,
        robust=robust,
        winsor_tail_p=winsor_tail_p,
    )
    out["s2_prior"] = out["var_prior"]
    out["s2_post"] = pd.Series(out["var_post"], index=fit.coefficients.index)
    del out["var_prior"]
    del out["var_post"]
    out["t"] = coefficients / stdev_unscaled / np.sqrt(out["s2_post"].values[:, None])
    df_total = df_residual + out["df_prior"]
    df_pooled = np.nansum(df_residual)
    df_total = np.minimum(df_total, df_pooled)
    out["df_total"] = df_total
    out["p_value"] = out["t"].apply(
        lambda x: 2 * scipy.stats.t.cdf(-np.abs(x), df=df_total), axis=0
    )

    # B-statistic
    var_prior_lim = np.array(stdev_coef_lim) ** 2 / np.median(out["s2_prior"])
    out["var_prior"] = tmixture_matrix(
        out["t"], stdev_unscaled, df_total, proportion, var_prior_lim
    )
    if np.any(np.isnan(out["var_prior"])):
        out["var_prior"][np.isnan(out["var_prior"])] = 1 / out["s2_prior"]
        logging.warnings.warn("Estimation of var_prior failed -- set to default value")

    r = np.outer(np.ones(out["t"].shape[0]), out["var_prior"])
    r = (stdev_unscaled**2 + r) / stdev_unscaled**2
    t2 = out["t"] ** 2
    Infdf = out["df_prior"] > 1e6
    if np.any(Infdf):
        kernel = t2 * (1 - 1 / r) / 2
        if np.any(~Infdf):
            t2_f = t2[~Infdf]
            r_f = r[~Infdf]
            df_total_f = df_total[~Infdf]
            kernel[~Infdf] = (
                (1 + df_total_f)
                / 2
                * np.log((t2_f + df_total_f) / (t2_f / r_f + df_total_f))
            )
    else:
        kernel = (
            (1 + df_total[:, None])
            / 2
            * np.log((t2 + df_total[:, None]) / (t2 / r + df_total[:, None]))
        )
    out["lods"] = np.log(proportion / (1 - proportion)) - np.log(r) / 2 + kernel
    return out


def tmixture_matrix(tstat, stdev_unscaled, df, proportion, v0_lim=None):
    """
    Estimate scale factor in mixture of *t*-distributions

    This function estimates the unscaled standard deviation of the true
    (unobserved) log fold changes for differentially expressed genes. It is
    used internally by :func:`eBayes` and is not intended to be called directly
    by users.

    The values in each column of :code:`tstat` are assumed to follow a mixture
    of an ordinary *t*-distribution, with mixing proportion
    :code:`1-proportion`, and :code:`(v0+v1)/v1` times a *t*-distribution, with
    mixing proportion :code:`proportion`. Here :code:`v1` is
    :code:`stdev_unscaled**2` and :code:`v0` is the value to be estimated.

    Arguments
    ---------
    tstat : ndarray
        matrix of *t*-statistics
    stdev_unscaled : ndarray
        matrix, conformal with :code:`tstat`, containing the unscaled standard
        deviations of the coefficients used to compute the *t*-statistics
    df : array_like
        array giving the degrees of freedom associated with :code:`tstat`
    proportion : float
        assumed proportion of genes that are differentially expressed
    v0_lim : pair of floats, optional
        lower and upper limits for the estimated unscaled standard deviations

    Returns
    -------
    ndarray
        estimated :code:`v0` values, one for each column of :code:`tstat`
    """
    if tstat.shape != stdev_unscaled.shape:
        raise ValueError("dimensions of tstat and stdev_unscaled do not match")
    if (tstat.shape[0],) != df.shape:
        raise ValueError("dimensions of tstat and df do not match")
    if v0_lim is not None:
        if len(v0_lim) != 2:
            raise ValueError("v0_lim must be a pair")
    ncoef = tstat.shape[1]
    if not isinstance(tstat, np.ndarray):
        tstat = np.array(tstat)
    if not isinstance(stdev_unscaled, np.ndarray):
        stdev_unscaled = np.array(stdev_unscaled)
    v0 = np.array(
        [
            tmixture_vector(tstat[:, j], stdev_unscaled[:, j], df, proportion, v0_lim)
            for j in range(ncoef)
        ]
    )
    return v0


def tmixture_vector(tstat, stdev_unscaled, df, proportion, v0_lim=None):
    """
    Estimate scale factor in mixture of *t*-distributions

    This function estimates the unscaled standard deviation of the true
    (unobserved) log fold changes for differentially expressed genes. It is
    used internally by :func:`eBayes` and is not intended to be called directly
    by users.

    The values in :code:`tstat` are assumed to follow a mixture of an ordinary
    *t*-distribution, with mixing proportion :code:`1-proportion`, and
    :code:`(v0+v1)/v1` times a *t*-distribution, with mixing proportion
    :code:`proportion`. Here :code:`v1` is :code:`stdev_unscaled**2` and
    :code:`v0` is the value to be estimated.

    Arguments
    ---------
    tstat : ndarray
        vector of *t*-statistics
    stdev_unscaled : ndarray
        vector, conformal with :code:`tstat`, containing the unscaled standard
        deviations of the coefficients used to compute the *t*-statistics
    df : array_like
        array giving the degrees of freedom associated with :code:`tstat`
    proportion : float
        assumed proportion of genes that are differentially expressed
    v0_lim : pair of floats, optional
        lower and upper limits for the estimated unscaled standard deviations

    Returns
    -------
    float
        estimated :code:`v0` value
    """
    if tstat.shape != stdev_unscaled.shape:
        raise ValueError("dimensions of tstat and stdev_unscaled do not match")
    if tstat.shape != df.shape:
        raise ValueError("dimensions of tstat and df do not match")
    tstat = np.asarray(tstat, dtype=float)
    stdev_unscaled = np.asarray(stdev_unscaled, dtype=float)
    # remove missing values
    if np.isnan(tstat).any():
        o = ~np.isnan(tstat)
        tstat = tstat[o]
        stdev_unscaled = stdev_unscaled[o]
        df = df[o]

    # ntarget t-statistics will be used for estimation
    ngenes = len(tstat)
    ntarget = int(np.ceil(proportion / 2 * ngenes))
    if ntarget < 1:
        return np.nan

    # if ntarget is very small, ensure p at least matches selected proportion
    # this ensures ptarget < 1
    p = np.max([ntarget / ngenes, proportion])

    # method requires that df be equal
    tstat = np.abs(tstat)
    MaxDF = np.max(df)
    i = df < MaxDF
    if np.any(i):
        # NB: original R code computes the proba on the log scale, but scipy
        # misses a "ilogsf" function
        TailP = scipy.stats.t.sf(tstat[i], df=df[i])
        tstat[i] = scipy.stats.t.isf(TailP, df=MaxDF)
        df[i] = MaxDF

    # Select top statistics
    o = np.flip(np.argsort(tstat))[:ntarget]
    tstat = tstat[o]
    v1 = stdev_unscaled[o] ** 2

    # compare to order statistics
    r = np.arange(ntarget) + 1.0
    p0 = 2 * scipy.stats.t.sf(tstat, df=MaxDF)
    ptarget = ((r - 0.5) / ngenes - (1 - p) * p0) / p
    v0 = np.zeros(ntarget)
    pos = ptarget > p0
    if pos.any():
        qtarget = scipy.stats.t.isf(ptarget[pos] / 2, df=MaxDF)
        v0[pos] = v1[pos] * ((tstat[pos] / qtarget) ** 2 - 1)
    if v0_lim is not None:
        v0 = np.clip(v0, v0_lim[0], v0_lim[1])
    return np.mean(v0)



import unittest

import pandas as pd

from inmoose.diffexp import DEResults


class Test(unittest.TestCase):
    def test_DEResults(self):
        df = pd.DataFrame()
        with self.assertRaisesRegex(
            ValueError, "log2FoldChange missing from results table"
        ):
            res = DEResults(df)

        df["log2FoldChange"] = [1, 2, 3]
        with self.assertRaisesRegex(ValueError, "lfcSE missing from results table"):
            res = DEResults(df)

        df["lfcSE"] = [0.1, 0.2, 0.3]
        with self.assertRaisesRegex(ValueError, "pvalue missing from results table"):
            res = DEResults(df)

        df["pvalue"] = [0.01, 0.02, 0.03]
        res = DEResults(df)

        _subres = res.loc[:1, :]



from .contrasts import contrasts_fit as contrasts_fit
from .contrasts import makeContrasts as makeContrasts
from .decidetests import TestResults as TestResults
from .decidetests import classifyTestsF as classifyTestsF
from .dups import uniquegenelist as uniquegenelist
from .dups import unwrapdups as unwrapdups
from .ebayes import eBayes as eBayes
from .ebayes import tmixture_matrix as tmixture_matrix
from .ebayes import tmixture_vector as tmixture_vector
from .fitFDist import fitFDist as fitFDist
from .lmfit import lm_series as lm_series
from .lmfit import lmFit as lmFit
from .lmfit import nonEstimable as nonEstimable
from .marraylm import MArrayLM as MArrayLM
from .squeezeVar import squeezeVar as squeezeVar
from .toptable import topTable as topTable



# -----------------------------------------------------------------------------
# Copyright (C) 2008-2022 Yunshun Chen, Aaron TL Lun, Davis J McCarthy, Matthew E Ritchie, Belinda Phipson, Yifang Hu, Xiaobei Zhou, Mark D Robinson, Gordon K Smyth
# Copyright (C) 2024 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# This file is based on the file 'R/binomTest.R' of the Bioconductor edgeR package (version 3.38.4).

import numpy as np
from scipy.stats import binom, chi2_contingency


def binomTest(y1, y2, n1=None, n2=None, p=None):
    """
    Computes *p*-values for differential abundance for each gene between two digital libraries, conditioning on the total counts for each gene. The counts in each group as a proportion of the whole are assumed to follow a binomial distribution.

    This function can be used to compare two libraries from SAGE, RAN-Seq,
    ChIP-Seq or other sequencing technologies with respect to technical
    variation.

    An exact two-sided binomial test is computed for each gene. This test is
    closely related to Fisher's exact test for 2x2 contingency tables but,
    unlike Fisher's test, it conditions on the total number of counts for each
    gene. The null hypothesis is that the expected counts are in the same
    proportions as the library sizes, *i.e.* that the binomial probability for
    the first library is :code:`n1 / (n1+n2)`.

    The two-sided rejection region is chosen analogously to Fisher's test.
    Specifically, the rejection region consists of those values with smallest
    probabilities under the null hypothesis.

    Then the counts are reasonably large, the binomial test, Fisher's test and
    Pearson's chi square all give the same results. When the counts are
    smaller, the binomial test is usually to be preferred in this context.

    References
    ----------
    `Binomial test <http://en.wikipedia.org/wiki/Binomial_test>`_
    `Fisher's exact test <http://en.wikipedia.org/wiki/Fisher's_exact_test>`_
    `Serial analysis of gene expression <http://en.wikipedia.org/wiki/Serial_analysis_of_gene_expression>`_

    Arguments
    ---------
    y1 : array_like
        integer array giving the count for each gene in the first library.
        Non-integer values are rounded to the nearest integer.
    y2 : array_like
        integer array giving the count for each gene in the second library.
        Non-integer values are rounded to the nearest integer.
    n1 : int
        total number of counts in the first library, across all genes. Not
        required if :code:`p` is supplied.
    n2 : int
        total number of counts in the second library, across all genes. Not
        required if :code:`p` is supplied.
    p : float
        expected proportion of :code:`y1` to the total for each gene under the
        null hypothesis.

    Returns
    -------
    ndarray
        array of *p*-values
    """
    if np.isnan(y1).any() or np.isnan(y2).any():
        raise ValueError("missing values not allowed")
    y1 = np.round(y1)
    y2 = np.round(y2)

    if y1.ndim > 1 or y2.ndim > 1:
        raise ValueError("y1 and y2 must be 1-D arrays")
    if len(y1) != len(y2):
        raise ValueError("y1 and y2 must have same length")
    if (y1 < 0).any() or (y2 < 0).any():
        raise ValueError("y1 and y2 must be non-negative")

    if n1 is None:
        n1 = y1.sum()
    if n2 is None:
        n2 = y2.sum()
    if p is None:
        p = n1 / (n1 + n2)

    if p <= 0 or p >= 1:
        raise ValueError("p must be between 0 and 1")
    size = y1 + y2
    p_value = np.ones(len(y1))
    if p == 0.5:
        i = size > 0
        if i.any():
            y1 = np.minimum(y1[i], y2[i])
            size = size[i]
            p_value[i] = np.minimum(2 * binom.cdf(y1, n=size, p=0.5), 1)
        return p_value
    big = size > 10000
    if big.any():
        for i in range(len(y1)):
            if big[i]:
                p_value[i] = chi2_contingency(
                    np.array([[y1[i], n1 - y1[i]], [y2[i], n2 - y2[i]]])
                ).pvalue
    size0 = size[(size > 0) & ~big]
    for isize in np.unique(size0):
        i = size == isize
        d = binom.pmf(np.arange(isize + 1), p=p, n=isize)
        o = np.argsort(d, kind="stable")
        cumsump = d[o].cumsum()[np.argsort(o, kind="stable")]
        p_value[i] = cumsump[y1[i]]
    return p_value



# -----------------------------------------------------------------------------
# Copyright (C) 2004-2022 Gordon Smyth, Yifang Hu, Matthew Ritchie, Jeremy Silver, James Wettenhall, Davis McCarthy, Di Wu, Wei Shi, Belinda Phipson, Aaron Lun, Natalie Thorne, Alicia Oshlack, Carolyn de Graaf, Yunshun Chen, Mette Langaas, Egil Ferkingstad, Marcus Davy, Francois Pepin, Dongseok Choi
# Copyright (C) 2024 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# This file is based on the file 'R/lmfit.R' of the Bioconductor limma package (version 3.55.1).

import logging

import numpy as np
import pandas as pd
import patsy
import scipy

from ..utils import lm_fit, lm_wfit
from .dups import uniquegenelist, unwrapdups
from .marraylm import MArrayLM


class EAWP:
    def __init__(self, obj):
        if not isinstance(obj, pd.DataFrame):
            self.exprs = pd.DataFrame(obj)
        else:
            self.exprs = obj.copy()
        self.design = None
        self.weights = None
        self.probes = None
        self.Amean = self.exprs.mean(axis=1)


def lmFit(
    obj,
    design=None,
    ndups=None,
    spacing=None,
    block=None,
    correlation=None,
    weights=None,
    method="ls",
):
    """
    Fit linear models for each gene given a series of arrays

    This function fits multiple linear models by weighted or generalized least
    squares. It accepts data from an experiment involving a series of
    microarrays with the same set of probes. A linear model is fitted to the
    expression data of each probe. The expression data should be log-ratios for
    two-color array platforms or log-expression values for one-channel
    platforms. To fit linear models to the individual channels of two-color
    array data, see :func:`lmscFit`. The coefficients of the fitted models
    describe the differences between the RNA sources hybridized to the arrays.
    The probe-wise fitted model results are stored in a compact form suitable
    for further processing by other functions in the limma package.

    This function allows for missing values and accepts quantitative precision
    weights through the :code:`weights` argument. It also supported two
    different correlation structures. If :code:`block` is not :code:`None`,
    then different arrays are assumed to be correlated. If :code:`block` is
    :code:`None` and :code:`ndups` is greater than one then replicate spots on
    the same array are assumed to be correlated. It is not possible at this
    time to fit models with a block structure and a duplicate-spot correlation
    structure simultaneously.

    If :code:`obj` is a matrix then it should contain log-ratios or
    log-expression data with rows corresponding to probes and columns to
    arrays. A vector is treated the same as a matrix with a single column. For
    objects of other classes, a matrix of expression values is taken from the
    appropriate component or slot of the object. If :code:`obj` is of class
    :code:`MAList` or :code:`marrayNorm`, then the matrix of log-ratios
    (M-values) is extracted.

    The arguments :code:`design`, :code:`ndups`, :code:`spacing` and
    :code:`weights` will be extracted from the data :code:`obj` if available.
    On the other hand, if any of these are set to a non-:code:`None` value in
    the function call then this value will override the value found in
    :code:`obj`.

    If the argument :code:`block` is used, then it is assumed that :code:`ndups=1`.

    The :code:`correlation` argument has a default value of :code:`0.75`, but
    in normal use this default value should not be relied on and the
    correlation value should be estimated using the function
    :code:`duplicateCorrelation`. The default value is likely to be too high in
    particular if used with the :code:`block` argument.

    The actual linear model computations are done by passing the data to one of
    the lower-level functions :func:`lm_series`, :func:`gls_series` or
    :func:`mrlm`. The function :func:`mrlm` is used if :code:`method="robust"`.
    If :code:`method="ls"`, then :func:`gls_series` is used if a correlation
    structure has been specified, *i.e.* if :code:`ndups > 1` or :code:`block`
    is non-null and :code:`correlation` is different from zero. If
    :code:`method="ls"` and there is no correlation structure,
    :code:`lm_series` is used.

    An overview of linear model functions in limma is given by :ref:`linearmodels`.

    See Also
    --------
    getEAWP
        extract expression values, gene annotation and so from the data :code:`obj`.

    Arguments
    ---------
    obj : matrix-like
        a matrix-like data object containing log-ratios or log-expression
        values for a series of arrays, with rows corresponding to genes and
        columns to samples. Any type of data object that can be processed by
        :func:`getEAWP` is acceptable.
    design : patsy formula-like
        the design matrix of the microarray experiment, with rows corresponding
        to samples and columns to coefficients to be estimated.
        Defaults to :code:`obj.design` if not :code:`None`, otherwise to the
        unit vector, meaning that all samples will be treated as replicated of
        a single treatment group.
    ndups : int
        positive integer giving the number of times each distinct probe is
        printed on each array
    spacing : int
        positive integer giving the spacing between duplicate occurrences of
        the same probe, :code:`spacing=1` for consecutive rows.
    block : array-like
        vector or factor specifying a blocking variable on the arrays. Has
        length equal to the number of arrays. Must be :code:`None` if
        :code:`ndups > 2`.
    correlation
        the inter-duplicate or inter-technical replicate correlation
    weights
        non-negative precision weights. Can be a numeric matrix of individual
        weights of same size as the object expression matrix, or a numeric
        vector of gene weights with length equal to :code:`nrow` of the
        expression matrix.
    method : { "ls", "robust" }
        fitting method: :code:`"ls"` for least squares or :code:`"robust"` for
        robust regression

    Returns
    -------
    MArrayLM
        object containing the result of the fits.
        The row names of :code:`obj` are preserved in the fit object and can be
        retrieved by :code:`fit.index` where :code:`fit` is the output of
        :func:`lmFit`.
        The column names of :code:`design` are preserved as column names and
        can be retrieved by :code:`fit.columns`.
    """
    # Fit genewise linear models

    # Extract components from obj
    y = EAWP(obj)

    if y.exprs.shape[0] == 0:
        raise ValueError("expression matrix has zero rows")

    # Check design matrix
    if design is None:
        design = y.design
    if design is None:
        design = patsy.dmatrix("~1", data=y.exprs)
    else:
        if not isinstance(design, patsy.DesignMatrix):
            raise ValueError("design must be a patsy DesignMatrix")
        if design.shape[0] != y.exprs.shape[1]:
            raise ValueError(
                "row dimension of design does not match column dimension of data object"
            )
        if np.isnan(design).any():
            raise ValueError("NAs not allowed in design matrix")

    ne = nonEstimable(design)
    if ne is not None:
        logging.info(f"Coefficients not estimable: {' '.join(ne)}")

    # Check ndups and spacing. Default to 1.
    if ndups is None:
        try:
            ndups = y.printer.ndups
        except AttributeError:
            pass
    if ndups is None:
        ndups = 1
    if spacing is None:
        try:
            spacing = y.printer.spacing
        except AttributeError:
            pass
    if spacing is None:
        spacing = 1

    # Check weights
    if weights is None:
        weights = y.weights

    # Check method
    if method not in ["ls", "robust"]:
        raise ValueError(f"unknown method: {method}")

    # If duplicates are present, reduce probe-annotation and Amean to correct length
    if ndups > 1:
        if y.probes is not None:
            y.probes = uniquegenelist(y.probes, ndups=ndups, spacing=spacing)
        if y.Amean is not None:
            y.Amean = np.nanmean(
                unwrapdups(y.Amean, ndups=ndups, spacing=spacing), axis=1
            )

    # Dispatch fitting algorithms
    if method == "robust":
        raise NotImplementedError("robust method for lmFit is not implemented yet")
        # fit = mrlm(
        #    y.exprs, design=design, ndups=ndups, spacing=spacing, weights=weights
        # )
    else:
        if ndups < 2 and block is None:
            fit = lm_series(
                y.exprs, design=design, ndups=ndups, spacing=spacing, weights=weights
            )
        else:
            if correlation is None:
                raise ValueError(
                    "the correlation must be set, see function duplicateCorrelation"
                )
            raise NotImplementedError("gls_series is not implemented yet")
            # fit = gls_series(
            #    y.exprs,
            #    design=design,
            #    ndups=ndups,
            #    spacing=spacing,
            #    block=block,
            #    correlation=correlation,
            #    weights=weights,
            # )

    # Possible warning on missing coefs
    if fit.coefficients.shape[1] > 1:
        n = np.sum(np.isnan(fit.coefficients), axis=1)
        n = np.sum((n > 0) & (n < fit.coefficients.shape[1]))
        if n > 0:
            logging.warnings.warn(f"Partial NA coefficients for {n} probes")

    # output
    fit.genes = y.probes
    fit.Amean = y.Amean
    fit.method = method
    fit.design = design
    return fit


def lm_series(M, design=None, ndups=1, spacing=1, weights=None):
    """
    Fit linear model to microarray data by ordinary least squares

    Fit a linear model genewise to expression data from a series of arrays.
    This function uses ordinary least squares and is a utility function for
    :func:`lmFit`. Most users should not use this function directly but should
    use :func:`lmFit` instead.

    The linear model is fit for each gene by calling the function
    :func:`~inmoose.utils.lm_fit` or :func:`~inmoose.utils.lm_wfit`.

    Arguments
    ---------
    M : pd.DataFrame
        numeric matrix containing log-ratio or log-expression values for a
        series of microarrays, rows correspond to genes and columns to arrays
    design
        design matrix defining the linear model. The number of rows should
        agree with the number of columns of M. The number of columns will
        determine the number of coefficients estimated for each gene.
    ndups : int
        number of duplicate spots. Each gene is printed :code:`ndups` times in
        adjacent spots on each array
    spacing : int
        the spacing between the rows of :code:`M` corresponding to duplicate
        spots, :code:`spacing=1` for consecutive spots.
    weights : array_like
        an optional matrix of the same dimension as :code:`M` containing
        weights for each spot. If it is of different dimension to :code:`M`, it
        will be filled out to the same size.

    Returns
    -------
    MArrayLM
        an object with attributes

        - coefficients: matrix containing the estimated coefficients for each
          linear model. Same number of rows as :code:`M`, same number of
          columns as :code:`design`.
        - stdev_unscaled: matrix conformal with :code:`coef` containing the
          unscaled standard deviations for the coefficient estimators. The
          standard errors are given by :code:`stdev_unscaled * sigma`.
        - sigma: numeric vector containing the residual standard deviation for
          each gene.
        - df_residual: vector giving the degrees of freedom corresponding to
          :code:`sigma`.
        - qr: QR-decomposition of :code:`design`

    """
    # Check design
    if design is None:
        design = patsy.dmatrix("~1", data=M)

    nbeta = design.shape[1]
    try:
        coef_names = design.design_info.column_names
    except AttributeError:
        coef_names = [f"x{i}" for i in range(nbeta)]

    # Check weights
    if weights is not None:
        weights = np.broadcast_to(weights, M.shape)
        weights[weights <= 0] = np.nan
        M[~np.isfinite(weights)] = np.nan

    # Reform duplicated rows into columns
    if ndups > 1:
        M = unwrapdups(M, ndups=ndups, spacing=spacing)
        design = design @ np.arange(1, ndups + 1)
        if weights is not None:
            weights = unwrapdups(weights, ndups=ndups, spacing=spacing)

    # Initialize standard errors
    ngenes = M.shape[0]
    stdev_unscaled = pd.DataFrame(
        np.full((ngenes, nbeta), np.nan), columns=coef_names, index=M.index
    )
    beta = pd.DataFrame(
        np.full((ngenes, nbeta), np.nan), columns=coef_names, index=M.index
    )

    # if QR-decomposition is constant for all genes, fit all genes in one sweep
    NoProbesWts = np.all(np.isfinite(M)) and (weights is None)
    if NoProbesWts:
        if weights is None:
            fit = lm_fit(design, M.T)
        else:
            fit = lm_wfit(design, M.T, weights[0, :])
            fit.weights = None

        if fit.df_residuals > 0:
            if fit.effects.ndim == 2:
                sigma = np.sqrt(np.mean(fit.effects[fit.rank :, :] ** 2, axis=0))
            else:
                sigma = np.sqrt(np.mean(fit.effects[fit.rank :] ** 2))
        else:
            sigma = np.full(ngenes, np.nan)

        cov_coef = np.linalg.inv(design.T @ design)
        cov_coef = pd.DataFrame(cov_coef, index=coef_names, columns=coef_names)
        stdev_unscaled.loc[:, coef_names] = np.sqrt(np.diag(cov_coef))
        coef = fit.coefficients.T
        coef.index = M.index
        coef.columns = coef_names
        return MArrayLM(
            coef,
            stdev_unscaled,
            pd.Series(sigma, index=M.index),
            np.full(ngenes, fit.df_residuals),
            cov_coef,
        )

    # Genewise QR-decompositions are required, so iterate through genes
    sigma = np.full(ngenes, np.nan)
    df_residual = np.zeros(ngenes)
    for i in range(ngenes):
        y = M.iloc[i, :]
        obs = np.isfinite(y)
        if obs.sum() > 0:
            X = design[obs, :]
            y = y[obs]
            if weights is None:
                out = lm_fit(X, y)
            else:
                w = weights[i, obs]
                out = lm_wfit(X, y, w)

            est = ~np.isnan(out.coefficients)
            beta[i, :] = out.coefficients
            stdev_unscaled[i, est] = np.sqrt(np.diag(np.linalg.inv(X.T @ X)))
            df_residual[i] = out.df_residuals
            if df_residual[i] > 0:
                sigma[i] = np.sqrt(np.mean(out.effects[out.rank :] ** 2))

    # Correlation matrix of coefficients
    cov_coef = np.linalg.inv(design.T @ design)

    return MArrayLM(
        coefficients=beta,
        stdev_unscaled=stdev_unscaled,
        sigma=sigma,
        df_residual=df_residual,
        cov_coef=cov_coef,
    )


def nonEstimable(x):
    """
    Check whether a matrix has full column rank

    Returns a vector of names for the columns of :code:`x` which are linearly
    dependent on previous columns. If :code:`x` has full column rank, then
    returns :code:`None`.

    Arguments
    ---------
    x : array_like
        a design matrix

    Returns
    -------
    list of strings or None
        list of the names of the matrix columns which are linearly dependent on
        previous columns. :code:`None` if :code:`x` has full column rank
    """
    p = x.shape[1]
    Q, R, pivot = scipy.linalg.qr(x, pivoting=True)
    rank = np.linalg.matrix_rank(R)
    if rank < p:
        try:
            n = x.design_info.column_names
        except AttributeError:
            n = [f"{i}" for i in range(p)]
        n = np.array(n)
        notest = n[pivot[rank:]]
        blank = notest == ""
        if np.any(blank):
            notest[blank] = np.array([f"{i}" for i in range(rank, p)])[blank]
        return notest
    else:
        return None
